# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Base custom error page
"""
import sys
from os.path import dirname, join

from ..exceptions import ActionRedirect, RequestBlocked
from ..frameworks.wsgi import WSGIResponse
from ..rules import RuleCallback

if sys.version_info >= (3, 0):
    from http.client import responses as HTTP_STATUS_CODE
else:
    from httplib import responses as HTTP_STATUS_CODE


class BaseSqreenErrorPage(RuleCallback):

    INTERRUPTIBLE = False
    DEFAULT_RULE_DATA = {
        "type": "custom_error_page",
        "status_code": 403,
    }

    def __init__(self, *args, **kwargs):
        super(BaseSqreenErrorPage, self).__init__(*args, **kwargs)
        with open(join(dirname(__file__), "sqreen_error_page.html"), "r") as f:
            self.content = f.read()

        rule_data = self.DEFAULT_RULE_DATA
        values = self.data.get("values")
        if values:
            rule_data = values[0]

        self.rule_type = rule_data.get("type", self.DEFAULT_RULE_DATA["type"])

        if self.rule_type == "custom_error_page":
            self.status_code = int(rule_data.get(
                "status_code", self.DEFAULT_RULE_DATA["status_code"]))
            self.storage.attack_http_code = self.status_code
        elif self.rule_type == "redirection" and "redirection_url" in rule_data:
            self.redirection_url = rule_data["redirection_url"]
        else:
            raise ValueError("Invalid rule_type %s" % self.rule_type)

    def handle_exception(self, exception):
        if exception is None or not isinstance(exception, RequestBlocked):
            return None

        exception.done()
        if isinstance(exception, ActionRedirect):
            headers = {"Location": exception.target_url}
            self.record_observation("http_code", "303", 1)
            return (303, "", headers)
        # Else, exception is either ActionBlock or AttackBlocked.
        elif self.rule_type == "custom_error_page":
            self.record_observation("http_code", str(self.status_code), 1)
            return (self.status_code, self.content, {})
        elif self.rule_type == "redirection":
            headers = {"Location": self.redirection_url}
            self.record_observation("http_code", "303", 1)
            return (303, "", headers)


class SqreenErrorPage(BaseSqreenErrorPage):

    def failing(self, instance, args, kwargs, exc_info=None, **options):
        ret = self.handle_exception(exc_info[1])
        if ret is not None:
            # Pylons tweak to bypass the StatusCodeRedirect middleware
            environ = args[0]
            environ["pylons.status_code_redirect"] = None

            status_code, body, headers = ret
            status_msg = HTTP_STATUS_CODE.get(status_code)
            if status_msg is not None:
                status = "{} {}".format(status_code, status_msg)
            else:
                status = "500 Internal Server Error"
            return {
                "status": "override",
                "new_return_value": WSGIResponse(status, headers, body.encode("utf-8"))
            }



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Reflected XSS callback
"""
import sys
from logging import getLogger

from ..utils import STRING_CLASS, UNICODE_CLASS, is_string
from .regexp_rule import RegexpRule

if sys.version_info >= (3, 2):
    # Added in Python 3.2
    from html import escape
else:
    # Removed in Python 3.8
    from cgi import escape


LOGGER = getLogger(__name__)


class ReflectedXSSCB(RegexpRule):
    def post(self, instance, args, kwargs, result=None, **options):
        """ Check if a template node returns a content that is in the
        query parameters
        """
        request = self.storage.get_current_request()

        if not request:
            LOGGER.warning("No request was recorded abort")
            return

        if not is_string(result):
            LOGGER.debug("Non string passed, type %s", type(result))
            return

        if request.params_contains(result):
            # If the payload is malicious, record the attack
            matching_regexp = self.match_regexp(result)
            if matching_regexp:
                self.record_attack(
                    {"found": matching_regexp, "payload": result}
                )

            # Only if the callback should block, sanitize the string
            if self.block:
                return {
                    "status": "override",
                    "new_return_value": self._escape(result),
                }

    @staticmethod
    def _escape(value):
        """ Escape a malicious value to make it safe
        """

        # Convert the value if it's a string subclass to bypass escape
        value_class = value.__class__
        if value_class not in (STRING_CLASS, UNICODE_CLASS):
            bases = value_class.__bases__

            if UNICODE_CLASS in bases:
                value = UNICODE_CLASS(value)
            elif STRING_CLASS in bases:
                value = STRING_CLASS(value)
            else:
                err_msg = "Value '{!r}' has invalid type and bases: {!r}"
                raise TypeError(err_msg.format(value, bases))

        # Convert to str before avoiding problems with Jinja2 Markup classes
        return escape(value, True)



# -*- coding: utf-8 -*-
#
#  Copyright 2011 Sybren A. Stüvel <sybren@stuvel.eu>
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""VARBLOCK file support

.. deprecated:: 3.4

    The VARBLOCK format is NOT recommended for general use, has been deprecated since
    Python-RSA 3.4, and will be removed in a future release. It's vulnerable to a
    number of attacks:

    1. decrypt/encrypt_bigfile() does not implement `Authenticated encryption`_ nor
       uses MACs to verify messages before decrypting public key encrypted messages.

    2. decrypt/encrypt_bigfile() does not use hybrid encryption (it uses plain RSA)
       and has no method for chaining, so block reordering is possible.

    See `issue #19 on Github`_ for more information.

.. _Authenticated encryption: https://en.wikipedia.org/wiki/Authenticated_encryption
.. _issue #19 on Github: https://github.com/sybrenstuvel/python-rsa/issues/13


The VARBLOCK file format is as follows, where || denotes byte concatenation:

    FILE := VERSION || BLOCK || BLOCK ...

    BLOCK := LENGTH || DATA

    LENGTH := varint-encoded length of the subsequent data. Varint comes from
    Google Protobuf, and encodes an integer into a variable number of bytes.
    Each byte uses the 7 lowest bits to encode the value. The highest bit set
    to 1 indicates the next byte is also part of the varint. The last byte will
    have this bit set to 0.

This file format is called the VARBLOCK format, in line with the varint format
used to denote the block sizes.

"""

import warnings

from ._compat import byte, b

ZERO_BYTE = b('\x00')
VARBLOCK_VERSION = 1

warnings.warn("The 'rsa.varblock' module was deprecated in Python-RSA version "
              "3.4 due to security issues in the VARBLOCK format. See "
              "https://github.com/sybrenstuvel/python-rsa/issues/13 for more information.",
              DeprecationWarning)


def read_varint(infile):
    """Reads a varint from the file.

    When the first byte to be read indicates EOF, (0, 0) is returned. When an
    EOF occurs when at least one byte has been read, an EOFError exception is
    raised.

    :param infile: the file-like object to read from. It should have a read()
        method.
    :returns: (varint, length), the read varint and the number of read bytes.
    """

    varint = 0
    read_bytes = 0

    while True:
        char = infile.read(1)
        if len(char) == 0:
            if read_bytes == 0:
                return 0, 0
            raise EOFError('EOF while reading varint, value is %i so far' %
                           varint)

        byte = ord(char)
        varint += (byte & 0x7F) << (7 * read_bytes)

        read_bytes += 1

        if not byte & 0x80:
            return varint, read_bytes


def write_varint(outfile, value):
    """Writes a varint to a file.

    :param outfile: the file-like object to write to. It should have a write()
        method.
    :returns: the number of written bytes.
    """

    # there is a big difference between 'write the value 0' (this case) and
    # 'there is nothing left to write' (the false-case of the while loop)

    if value == 0:
        outfile.write(ZERO_BYTE)
        return 1

    written_bytes = 0
    while value > 0:
        to_write = value & 0x7f
        value >>= 7

        if value > 0:
            to_write |= 0x80

        outfile.write(byte(to_write))
        written_bytes += 1

    return written_bytes


def yield_varblocks(infile):
    """Generator, yields each block in the input file.

    :param infile: file to read, is expected to have the VARBLOCK format as
        described in the module's docstring.
    @yields the contents of each block.
    """

    # Check the version number
    first_char = infile.read(1)
    if len(first_char) == 0:
        raise EOFError('Unable to read VARBLOCK version number')

    version = ord(first_char)
    if version != VARBLOCK_VERSION:
        raise ValueError('VARBLOCK version %i not supported' % version)

    while True:
        (block_size, read_bytes) = read_varint(infile)

        # EOF at block boundary, that's fine.
        if read_bytes == 0 and block_size == 0:
            break

        block = infile.read(block_size)

        read_size = len(block)
        if read_size != block_size:
            raise EOFError('Block size is %i, but could read only %i bytes' %
                           (block_size, read_size))

        yield block


def yield_fixedblocks(infile, blocksize):
    """Generator, yields each block of ``blocksize`` bytes in the input file.

    :param infile: file to read and separate in blocks.
    :returns: a generator that yields the contents of each block
    """

    while True:
        block = infile.read(blocksize)

        read_bytes = len(block)
        if read_bytes == 0:
            break

        yield block

        if read_bytes < blocksize:
            break



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
"""Wrapper class for aiohttp request objects."""

from logging import getLogger

from ..utils import to_unicode_safe
from .base import BaseRequest, BaseResponse

LOGGER = getLogger(__name__)


def multidict_to_dict(multidict):
    """Convert a multidict instance to a regular dict.

    Each key in the resulting dict is associated to the list of values mapped
    to this key in the original multidict.
    """
    getter = getattr(multidict, "getall", multidict.get)
    return {key: getter(key) for key in multidict}


def get_request_remote(request):
    """Get the remote IP address of an aiohttp.web.Request object.

    Return a string with the remote IP address if it could be determined, None
    otherwise.

    This function calls the property request.remote if it is defined
    (introduced in aiohttp 2.3). If not, it mimics aiohttp implementation and
    fallbacks on socket.getpeername.
    """
    if hasattr(request, "remote"):
        return request.remote
    peername = request.transport.get_extra_info("peername")
    if isinstance(peername, (list, tuple)):
        return peername[0]
    else:
        return peername


class AioHTTPRequest(BaseRequest):
    """Wrapper class for aiohttp request objects."""

    def __init__(self, request):
        super(AioHTTPRequest, self).__init__()
        self.request = request

        # Caches for lazy properties.
        self._headers = None
        self._query_params = None
        self._form_params = None
        self._cookies_params = None
        self._json_params = None

    @property
    def server_port(self):
        """Server port number."""
        sockname = self.request.transport.get_extra_info("sockname")
        return sockname[1]

    @property
    def remote_port(self):
        """Client port number."""
        peername = self.request.transport.get_extra_info("peername")
        return peername[1]

    @property
    def method(self):
        """Request method."""
        return self.request.method

    @property
    def scheme(self):
        """Request scheme (http or https)."""
        return self.request.scheme

    @property
    def hostname(self):
        """Request host."""
        return self.request.host

    @property
    def route(self):
        """Request route."""
        match_info = getattr(self.request, "match_info", None)
        if match_info is None:
            return None
        match_info = match_info.get_info()
        if "path" in match_info:
            # Static route (PlainResource).
            return match_info["path"]
        elif "formatter" in match_info:
            # Pattern-based route (DynamicResource).
            return match_info["formatter"]
        else:
            # HTTP exceptions, e.g. 404.
            return None

    @property
    def path(self):
        """Request path."""
        return self.request.path

    @property
    def request_uri(self):
        """Request URI."""
        return self.request.rel_url.path_qs

    @property
    def raw_headers(self):
        """Dictionary of request headers.

        Headers are encoded in WSGI format, described in PEP 3333.
        """
        if self._headers is None:
            headers = {
                "REQUEST_METHOD": self.request.method,
                "SCRIPT_NAME": "",  # Mandatory as per PEP 3333.
                "PATH_INFO": self.request.path,
                "QUERY_STRING": self.request.query_string,
                "SERVER_NAME": self.hostname,
                "SERVER_PORT": str(self.server_port),
                "SERVER_PROTOCOL": "HTTP/{}.{}".format(*self.request.version),
                "wsgi.scheme": self.request.scheme,
            }
            remote = get_request_remote(self.request)
            if remote is not None:
                headers["REMOTE_ADDR"] = remote
            if self.request.content_length is not None:
                headers["CONTENT_LENGTH"] = str(self.request.content_length)
            if self.request.content_type is not None:
                headers["CONTENT_TYPE"] = self.request.content_type
            for key, value in self.request.headers.items():
                name = "HTTP_{}".format(key.replace("-", "_").upper())
                headers[name] = value
            self._headers = headers
        return self._headers

    @property
    def referer(self):
        """Request referer."""
        return to_unicode_safe(self.get_raw_header("HTTP_REFERER"))

    @property
    def client_user_agent(self):
        """User agent."""
        return to_unicode_safe(self.get_raw_header("HTTP_USER_AGENT"))

    @property
    def remote_addr(self):
        """Remote IP address."""
        return to_unicode_safe(self.get_raw_header("REMOTE_ADDR"))

    @property
    def query_params(self):
        """Dictionary of query parameters."""
        if self._query_params is None:
            self._query_params = multidict_to_dict(self.request.query)
        return self._query_params

    @property
    def form_params(self):
        """Dictionary of form parameters."""
        if self._form_params is None:
            from .async_utils import run_coroutine

            post_params = run_coroutine(self.request.post())
            self._form_params = multidict_to_dict(post_params)
        return self._form_params

    @property
    def cookies_params(self):
        """Dictionary of cookies parameters."""
        if self._cookies_params is None:
            self._cookies_params = multidict_to_dict(self.request.cookies)
        return self._cookies_params

    @property
    def json_params(self):
        """Request body, decoded as JSON."""
        if self._json_params is None:
            try:
                from .async_utils import run_coroutine

                json_params = run_coroutine(self.request.json())
                self._json_params = multidict_to_dict(json_params)
            except Exception:
                LOGGER.debug("couldn't get request.json from the framework",
                             exc_info=True)
                self._json_params = {}
        return self._json_params

    @property
    def view_params(self):
        """Dictionary of view arguments that matched the request."""
        return self.request.match_info


class AioHTTPResponse(BaseResponse):

    def __init__(self, response):
        self.response = response

    @property
    def status_code(self):
        return self.response.status

    @property
    def content_type(self):
        return self.response.headers.get("Content-Type")

    @property
    def content_length(self):
        try:
            return int(self.response.content_length)
        except (ValueError, TypeError):
            return None



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Contains all callback classes
"""
import logging

from ..exceptions import InvalidSignature
from ..remote_exception import RemoteException
from ..rules import RuleCallback
from ..runtime_storage import runtime
from ..signature import REQUIRED_KEYS, SIGNATURE_VERSION
from ..utils import HAS_ASYNCIO
from .auth_metrics import AuthMetricsCB
from .binding_accessor_counter import BindingAccessorCounter
from .binding_accessor_matcher import BindingAccessorMatcherCallback
from .count_http_codes import CountHTTPCodesCB
from .count_http_codes_django import CountHTTPCodesCBDjango
from .count_http_codes_flask import CountHTTPCodesCBFlask
from .count_http_codes_php import CountHTTPCodesCBPHP
from .count_http_codes_pyramid import CountHTTPCodesCBPyramid
from .crawler_user_agent_matches_metrics import (
    CrawlerUserAgentMatchesMetricsCB,
)
from .extension_passthru import ExtensionPassthruCB
from .headers_insert import HeadersInsertCB
from .headers_insert_django import HeadersInsertCBDjango
from .headers_insert_flask import HeadersInsertCBFlask
from .headers_insert_pyramid import HeadersInsertCBPyramid
from .ip_action import IPActionCB
from .ip_blacklist import IPBlacklistCB
from .not_found import NotFoundCB
from .not_found_django import NotFoundCBDjango
from .not_found_flask import NotFoundCBFlask
# Import rules callbacks
from .record_request import RecordRequest
from .record_request_context import RecordRequestContext
from .record_request_context_django import RecordRequestContextDjango
from .record_request_context_flask import RecordRequestContextFlask
from .record_request_context_pyramid import RecordRequestContextPyramid
from .record_request_django import RecordRequestDjango
from .record_response import RecordResponse
from .reflected_xss import ReflectedXSSCB
from .regexp_rule import RegexpRule
from .shell_env import ShellEnvCB
from .sqreen_error_page import SqreenErrorPage
from .sqreen_error_page_django import SqreenErrorPageDjango
from .sqreen_error_page_django_new_style import SqreenErrorPageDjangoNewStyle
from .sqreen_error_page_flask import SqreenErrorPageFlask
from .sqreen_error_page_pyramid import SqreenErrorPagePyramid
from .user_agent_matches import UserAgentMatchesCB
from .user_agent_matches_framework import (
    UserAgentMatchesCBDjango,
    UserAgentMatchesCBFramework,
)
from .waf import WAFCB
from .wsgi_input_preview import WSGIInputPreview

# Workaround for Jenkins build.
try:
    from .js import JSCB
except ImportError:
    pass

if HAS_ASYNCIO:
    from .record_request_context_aiohttp import RecordRequestContextAioHTTP
    from .headers_insert_aiohttp import HeadersInsertCBAioHTTP
    from .count_http_codes_aiohttp import CountHTTPCodesCBAioHTTP
    from .sqreen_error_page_aiohttp import SqreenErrorPageAioHTTP


LOGGER = logging.getLogger(__name__)


def cb_from_rule(rule, runner, rule_verifier=None, storage=runtime):
    """ Instantiate the right cb class
    """
    # Fixme clean

    if not isinstance(rule, dict):
        LOGGER.debug("Invalid rule type %s", type(rule))
        return

    # Check rule signature
    if rule_verifier is not None \
            and not rule_verifier.verify_rule(rule, REQUIRED_KEYS, SIGNATURE_VERSION):
        LOGGER.critical("Can't verify signature of rule %s", rule["name"])
        raise InvalidSignature(rule["name"], rule["rulespack_id"])

    # Ask for native attack rendering
    if "attack_render_info" in rule.get("hookpoint", {}):
        callback_class_name = "ExtensionPassthruCB"

    # Use JSCB for internal_js_hook callback used in PHP
    elif rule.get("hookpoint", {}).get("strategy") == "internal_js_hook":
        callback_class_name = "JSCB"

    else:
        callback_class_name = rule.get("hookpoint", {}).get("callback_class")

    if callback_class_name is None:
        err_msg = (
            "Couldn't find a callback_class_name for rule %s, fallback on js"
        )
        LOGGER.debug(err_msg, rule["name"])
        callback_class_name = "JSCB"

    possible_subclass = globals().get(callback_class_name, None)

    if possible_subclass and issubclass(possible_subclass, RuleCallback):
        try:
            return possible_subclass.from_rule_dict(rule, runner, storage=storage)
        except Exception:
            LOGGER.warning(
                "Couldn't instantiate a callback for rule %s",
                rule,
                exc_info=True,
            )
            infos = {
                "rule_name": rule["name"],
                "rulespack_id": rule["rulespack_id"],
            }
            remote_exception = RemoteException.from_exc_info(callback_payload=infos)
            runner.queue.put(remote_exception)
            return

    LOGGER.debug(
        "Couldn't find the class matching class_name %s", callback_class_name
    )
    return


__all__ = [
    "AuthMetricsCB",
    "BindingAccessorCounter",
    "BindingAccessorMatcherCallback",
    "cb_from_rule",
    "CountHTTPCodesCB",
    "CountHTTPCodesCBAioHTTP",
    "CountHTTPCodesCBDjango",
    "CountHTTPCodesCBFlask",
    "CountHTTPCodesCBPHP",
    "CountHTTPCodesCBPyramid",
    "CrawlerUserAgentMatchesMetricsCB",
    "ExtensionPassthruCB",
    "HeadersInsertCB",
    "HeadersInsertCBAioHTTP",
    "HeadersInsertCBDjango",
    "HeadersInsertCBFlask",
    "HeadersInsertCBPyramid",
    "IPActionCB",
    "IPBlacklistCB",
    "JSCB",
    "NotFoundCB",
    "NotFoundCBDjango",
    "NotFoundCBFlask",
    "RecordRequest",
    "RecordRequestDjango",
    "RecordRequestContext",
    "RecordRequestContextAioHTTP",
    "RecordRequestContextDjango",
    "RecordRequestContextFlask",
    "RecordRequestContextPyramid",
    "RecordResponse",
    "ReflectedXSSCB",
    "RegexpRule",
    "ShellEnvCB",
    "SqreenErrorPage",
    "SqreenErrorPageAioHTTP",
    "SqreenErrorPageDjango",
    "SqreenErrorPageDjangoNewStyle",
    "SqreenErrorPageFlask",
    "SqreenErrorPagePyramid",
    "UserAgentMatchesCB",
    "UserAgentMatchesCBDjango",
    "UserAgentMatchesCBFramework",
    "WAFCB",
    "WSGIInputPreview",
]



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
"""Helper classes for runtime storage."""

import threading

import pkg_resources

from .frameworks.blank import BlankRequest, BlankResponse
from .payload_creator import PayloadCreator
from .request_recorder import RequestRecorder
from .utils import HAS_ASYNCIO


class RuntimeStorage(object):
    """Runtime storage objects."""

    default_request_class = BlankRequest
    default_response_class = BlankResponse

    def __init__(self, store=None):
        if store is None:
            store = _ThreadLocalStore()
        self.store = store
        # Let's not set store attributes here: the event loop may not be
        # started yet.
        self.payload_creator = PayloadCreator()
        self.attack_http_code = 403

    def store_request(self, request, **kwargs):
        """Store a request."""
        self.store.set("current_request", request)
        request_recorder = self.store.get("request_recorder")
        if request_recorder is None:
            # First request.
            request_recorder = RequestRecorder()
            self.store.set("request_recorder", request_recorder)
        request_recorder.start_request(request, **kwargs)

    def store_request_default(self, request, **kwargs):
        """Store a request, unless a request is already stored."""
        current_request = self.get_current_request()
        if current_request is None:
            self.store_request(request, **kwargs)

    def get_current_request(self):
        """Return the request currently processed.

        Return None if no request is stored.
        """
        return self.store.get("current_request")

    def clear_request(self, queue=None, observation_queue=None):
        """Clear the stored request."""
        self.store.delete("whitelist_match")
        self.store.delete("current_request_input_preview")
        current_request = self.store.delete("current_request")
        current_response = self.store.delete("current_response")
        request_recorder = self.store.get("request_recorder")
        if request_recorder is not None:
            request_recorder.end_request()
            if current_request is not None and queue is not None \
                    and observation_queue is not None:
                request_recorder.flush(
                    current_request, current_response,
                    self.payload_creator, queue, observation_queue
                )
            else:
                request_recorder.clear()

    def store_request_input_preview(self, preview, end_of_input):
        """Store the request input stream body preview."""
        self.store.set("current_request_input_preview",
                       (preview, end_of_input))

    def get_request_input_preview(self):
        """Get the request input stream body preview."""
        return self.store.get("current_request_input_preview", (None, None))

    def clear_request_input_preview(self):
        """Clear the request input stream body preview."""
        self.store.delete("current_request_input_preview")

    def store_response(self, response):
        """Store a response."""
        self.store.set("current_response", response)

    def store_response_default(self, response):
        """Store a response, unless a response is already stored."""
        current_response = self.get_current_response()
        if current_response is None:
            self.store_response(response)

    def get_current_response(self):
        """Returns the current response or None."""
        return self.store.get("current_response")

    def clear_response(self, response):
        """Clear the stored response."""
        self.store.delete("current_response")

    @property
    def request_sq_time(self):
        """Proxy to RequestRecorder request_sq_time property."""
        request_recorder = self.store.get("request_recorder")
        if request_recorder is None:
            request_recorder = RequestRecorder()
            self.store.set("request_recorder", request_recorder)
        return request_recorder.request_sq_time

    def trace(self, *args, **kwargs):
        """Proxy to RequestRecorder trace method."""
        request_recorder = self.store.get("request_recorder")
        if request_recorder is None:
            request_recorder = RequestRecorder()
            self.store.set("request_recorder", request_recorder)
        return request_recorder.trace(*args, **kwargs)

    def observe(self, *args, **kwargs):
        """Proxy to RequestRecorder observe method."""
        request_recorder = self.store.get("request_recorder")
        if request_recorder is None:
            request_recorder = RequestRecorder()
            self.store.set("request_recorder", request_recorder)
        request_recorder.observe(*args, **kwargs)

    def get_whitelist_match(self):
        """Return the whitelisted path or IP matching the current request.

        Return False if the request is not whitelisted.
        """
        return self.store.get("whitelist_match", None)

    def set_whitelist_match(self, match):
        """Store a whitelisted path or IP matching the current request.
        """
        self.store.set("whitelist_match", match)

    def store_current_args(self, binding_accessor):
        """For daemon only: store the faked list of arguments."""
        args = {}
        for name, value in binding_accessor.items():
            if name.find("#.args[") == 0:
                idx = int(name[7 : name.rindex("]")])
                args[idx] = value
            elif name.find("#.cargs[") == 0:
                idx = int(name[8 : name.rindex("]")])
                args[idx] = value
        if args:
            self.store.set(
                "arguments", [args.get(i) for i in range(max(args) + 1)]
            )
        else:
            self.store.set("arguments", [])

    def get_current_args(self, args=None):
        """For daemon only: get the faked list of arguments."""
        return self.store.get("arguments", args)

    def store_cmd_arguments(self, cmd_args):
        """Alternative of store_current_args that doesn't destroy data"""
        self.store.set("cmd_args", cmd_args)

    def get_cmd_arguments(self):
        """dict ba expr -> its resolution by the extension"""
        return self.store.get("cmd_args")

    def store_extension_data(self, key, value):
        """Store extension data associated to a key."""
        extension_data = self.store.get("extension_data")
        if extension_data is None:
            extension_data = {}
            self.store.set("extension_data", extension_data)
        extension_data[key] = value

    def get_extension_data(self, key):
        """Returns the extension data associated to key or None."""
        return self.store.get("extension_data", {}).get(key)

    def clear_extension_data(self):
        """Clear the extension data."""
        self.store.delete("extension_data")


class _ThreadLocalStore(object):
    """Wrap a threading.local object into a store interface."""

    def __init__(self):
        self.local = threading.local()

    def get(self, key, default=None):
        """Return the value associated to the key.

        Fallback to default if the key is missing.
        """
        return getattr(self.local, key, default)

    def set(self, key, value):
        """Map a key to a value."""
        setattr(self.local, key, value)

    def setdefault(self, key, value):
        """If a key is not set, map it to a value."""
        self.local.__dict__.setdefault(key, value)

    def delete(self, key):
        """Delete a key."""
        return self.local.__dict__.pop(key, None)


_THREADED_FRAMEWORKS = ("Django", "Flask", "pyramid")

_ASYNC_FRAMEWORKS = ("aiohttp",)


def _get_store():
    """Return a store object suited for the current framework."""
    if HAS_ASYNCIO:
        pkg_names = set(
            pkg_info.project_name for pkg_info in pkg_resources.working_set
        )
        for framework in _THREADED_FRAMEWORKS:
            if framework in pkg_names:
                return _ThreadLocalStore()
        for framework in _ASYNC_FRAMEWORKS:
            if framework in pkg_names:
                from . import async_context

                return async_context
    return _ThreadLocalStore()


def get_runtime_storage():
    """Return a runtime storage instance suited for the current framework."""
    return RuntimeStorage(_get_store())


runtime = get_runtime_storage()



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#

import logging

from sqreen.exceptions import InvalidArgument
from sqreen.rules import RuleCallback
from sqreen.utils import Mapping, is_string

LOGGER = logging.getLogger(__name__)


class ExtensionPassthruCB(RuleCallback):
    """Daemon callback that just raises with the data the extension gives"""

    INTERRUPTIBLE = False

    def __init__(self, *args, **kwargs):
        super(ExtensionPassthruCB, self).__init__(*args, **kwargs)

        if not self.attack_render_info or not isinstance(self.attack_render_info, Mapping):
            msg = "Invalid data type received for 'attack_render_info': {}"
            raise InvalidArgument(msg.format(type(self.attack_render_info)))

    # info keys are stored like this in the rule:
    # rule = {
    #   "hookpoint": {
    #     "attack_render_info": {
    #       "pre": { <- upon data receiving, will generate attack with this info
    #         "found": "#.cargs[0]"
    #       }
    #     }
    #   }
    #   "callbacks": { <- determines what gets sent to the daemon
    #     "pre": [
    #       "#.request_params",
    #       "#.cargs[0]",
    #       "javascript code" <- extension ignores last element
    #     ]
    #   }
    # }
    def __fetch_render_info(self, cb_name):
        ret = self.attack_render_info.get(cb_name)
        if not isinstance(ret, dict) or \
                any([not is_string(k) or not is_string(v)
                     for (k, v) in ret.items()]):
            LOGGER.warning("Invalid attack rendering infos spec: %s", ret)
            return {}
        return ret

    def __build_info(self, attack_render_info):
        # map ba expr -> its resolution, as calculated by ext
        cmd_args = self.storage.get_cmd_arguments()
        if cmd_args is None:
            cmd_args = {}

        return {
            k: cmd_args.get(ba_expr) for (k, ba_expr) in attack_render_info.items()
        }

    def pre(self, instance, args, kwargs, **options):
        attack_render_info = self.__fetch_render_info('pre')
        attack_rec = self.__build_info(attack_render_info)

        status = attack_rec.pop("__status_override", "raise")
        if len(attack_rec) != 0:
            self.record_attack(attack_rec)

        return {'status': status}

    def post(self, instance, args, kwargs, **options):
        LOGGER.error("Called unsupported 'post' method")
        return

    def failing(self, instance, args, kwargs, **options):
        LOGGER.error("Called unsupported 'failing' method")
        return



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Execute and sanitize remote commands
"""
import json
import logging
from copy import copy
from itertools import groupby
from operator import itemgetter

from sqreen.performance_metrics import PerformanceMetricsSettings

from . import config
from .actions import ACTION_STORE
from .rules_callbacks import cb_from_rule
from .runtime_storage import runtime
from .sdk.events import STACKTRACE_EVENTS
from .signature import RSAVerifier

LOGGER = logging.getLogger(__name__)


class RemoteCommand(object):
    """ Class responsible for dispatching and executing remote commands
    """

    def __init__(self):
        self.commands = {}
        self.coalesce = set()

    def register_command(self, command_name, command, coalesce=False):
        """ Register a command callback for command name
        """
        self.commands[command_name] = command
        if coalesce:
            self.coalesce.add(command_name)

    def process_list(self, commands, *args, **kwargs):
        """ Process a list of command and assemble the result
        """
        res = {}
        if not isinstance(commands, list):
            LOGGER.debug(
                "Wrong commands type %s: %r", type(commands), commands
            )
            return res

        it = self.coalesce_commands(self.validate_commands(commands, res), res)
        for command in it:
            LOGGER.debug("Processing command %s", command["name"])

            # Command params
            command_params = command.get("params", [])

            # Then execute the command
            result = self.commands[command["name"]](
                *args, params=command_params, **kwargs
            )
            res[command["uuid"]] = self._format_result(result)

        return res

    def validate_commands(self, commands, results):
        """ Generator checking the commands have been registered.
        """
        for command in commands:
            uuid = command.get("uuid")
            if uuid is None:
                LOGGER.debug("skipping command without uuid: %r", command)
                continue
            name = command.get("name")
            if name not in self.commands:
                results[uuid] = {
                    "status": False,
                    "msg": "unknown command name {}".format(repr(name))
                }
                continue
            yield command

    def coalesce_commands(self, commands, results):
        """ Generator removing all duplicate adjacent commands.
        """
        for command_name, command_group in groupby(commands, itemgetter("name")):
            if command_name in self.coalesce:
                command_group = list(command_group)
                # Only keep the last command
                while len(command_group) > 1:
                    command = command_group.pop(0)
                    results[command["uuid"]] = {
                        "status": False,
                        "reason": "skipped duplicate {}".format(command_name)
                    }
            for command in command_group:
                yield command

    def process(self, command, *args, **kwargs):
        """ Process a single command (deprecated, only used by the PHP Daemon).
        """
        return self.process_list([command], *args, **kwargs).popitem()[1]

    @staticmethod
    def _format_result(result):
        """ Format the command result for the backend
        """
        if result is None:
            return {"status": False, "output": "None returned"}
        elif result is True:
            return {"status": True}
        else:
            return {"status": True, "output": result}

    @classmethod
    def with_production_commands(cls):
        """ Returns a RemoteCommand with all production commands
        already registered
        """
        remote_command = cls()

        for name, command in ALL_COMMANDS.items():
            remote_command.register_command(name, command,
                                            coalesce=name in COALESCE_COMMANDS)

        return remote_command


###
# COMMANDS DEFINITION
###


def _load_local_rules():
    config_value = config.CONFIG["RULES"]
    if not config_value:
        return []
    with open(config_value) as rules_file:
        local_rules = json.load(rules_file)
    if not isinstance(local_rules, list):  # Single rule.
        local_rules = [local_rules]
    for rule_dict in local_rules:
        rule_dict["rulespack_id"] = "local"
    return local_rules


def _load_rules(runner, params=None, check_signature=True, storage=runtime):
    """ Retrieve the rulespack and instantiate the callbacks, returns
    a list of callbacks
    """
    rulespack_id, rules = None, None
    # Try to load rules and rulespack from params
    if params:
        try:
            rulespack_id, rules = params
        except ValueError:
            pass

    if rulespack_id is None or rules is None:
        rulespack = runner.session.get_rulespack()
        if rulespack is not None:
            rulespack_id = rulespack.get("pack_id")
            rules = rulespack.get("rules", [])

    if rules is None or len(rules) == 0 or rulespack_id is None:
        return None, []

    # Set the pack id on each rule
    for rule_dict in rules:
        rule_dict["rulespack_id"] = rulespack_id

    rules.extend(_load_local_rules())

    LOGGER.info("Retrieved rulespack id: %s", rulespack_id)
    rules_name = ", ".join(rule["name"] for rule in rules)
    LOGGER.debug("Retrieved %d rules: %s", len(rules), rules_name)

    # Check the rule signature only if the config say so
    if check_signature and config.CONFIG["RULES_SIGNATURE"] is True:
        verifier = RSAVerifier()
    else:
        verifier = None

    # Create the callbacks
    callbacks = []
    for rule_dict in rules:
        # Instantiate the rule callback
        callback = cb_from_rule(rule_dict, runner, verifier, storage=storage)

        # Check if the rule has some metrics to register
        for metric in rule_dict.get("metrics", []):
            runner.metrics_store.register_metric(**metric)

        if callback:
            LOGGER.debug(
                'Rule "%s" will hook on "%s %s" with callback %s with strategy %s',
                callback.rule_name,
                callback.hook_module,
                callback.hook_name,
                callback,
                callback.strategy,
            )

            callbacks.append(callback)

    return rulespack_id, callbacks


def _instrument_callbacks(runner, callbacks):
    """ For a given list of callbacks, hook them
    """
    # Clean existing callbacks to avoid double-instrumentation
    instrumentation_remove(runner)
    LOGGER.info("Setup instrumentation")
    for callback in callbacks:
        runner.instrumentation.add_callback(callback)
    runner.instrumentation.hook_all()


def instrumentation_enable(runner, params=None):
    """ Retrieve a rulespack, instantiate RuleCallback from the rules
    and instrument them.
    """
    # Load rules
    pack_id, rules = _load_rules(runner, params)

    if pack_id is None:
        LOGGER.debug("Couldn't get a valid rulespack, instrumentation is disabled")
        return None

    LOGGER.debug("Start instrumentation with rulespack %r", pack_id)
    # Instrument retrieved rules
    _instrument_callbacks(runner, rules)

    return pack_id


def instrumentation_remove(runner, params=None):
    """ Remove all callbacks from instrumentation, return True
    """
    LOGGER.info("Remove current instrumentation")
    runner.instrumentation.deinstrument_all()
    return True


def get_bundle(runner, params=None):
    """ Returns the list of installed Python dependencies
    """

    return {"status": runner.session.post_bundle(runner.runtime_infos) is not None}


def rules_reload(runner, params=None):
    """ Load the rules, deinstrument the old ones and instrument
    the new loaded rules. Returns the new rulespack_id
    """
    pack_id, rules = _load_rules(runner)
    if pack_id is None:
        LOGGER.debug("Couldn't get a valid rulespack, instrumentation not reloaded")
        return ""

    LOGGER.debug("Reloading instrumentation with rulespack %r", pack_id)
    # Reinstrument with the new rules
    _instrument_callbacks(runner, rules)

    return pack_id


def actions_reload(runner, params=None):
    """Remove old security actions and load the new ones."""
    LOGGER.debug("Reloading actions")
    data = runner.session.get_actionspack()
    if data["status"] is not True:
        LOGGER.debug("Cannot reload actions")
        return ""

    actions = data["actions"]
    LOGGER.debug("Reloading actions %s", actions)

    unsupported = ACTION_STORE.reload_from_dicts(actions or [])
    if unsupported:
        return {
            "status": False,
            "output": {"unsupported_actions": unsupported},
        }
    else:
        return {"status": True}


def record_stacktrace(runner, params=None):
    """Collect stacktraces for the given events."""
    STACKTRACE_EVENTS.clear()
    if params:
        LOGGER.debug("Collect stacktraces for events %r", params)
        STACKTRACE_EVENTS.update(params)
    else:
        LOGGER.debug("Don't collect any stacktrace event")
    return {"status": True}


def features_get(runner, params=None):
    return runner.features_get()


def features_change(runner, params):

    old_features = copy(runner.features_get())

    if len(params) != 1:
        raise ValueError("Invalid params list %s" % params)
    params = params[0]

    runner.set_performance_metrics_settings(
        PerformanceMetricsSettings.from_features(params))

    # Heartbeat delay only if we have a new value for it > 0
    heartbeat_delay = params.get("heartbeat_delay", 0)
    if heartbeat_delay > 0:
        runner.set_heartbeat_delay(heartbeat_delay)

    # Always set the new value of call_counts_metrics_period
    runner.set_call_counts_metrics_period(
        params.get("call_counts_metrics_period", 60)
    )

    # Always set the new value of whitelisted_metric
    runner.set_whitelisted_metric(params.get("whitelisted_metric", True))

    # Alter the deliverer if the batch_size is set
    if params.get("batch_size") is not None:
        runner.set_deliverer(
            params["batch_size"], params.get("max_staleness", 0),
            use_signals=params.get("use_signals", False)
        )

    new_features = copy(runner.features_get())

    return {"old": old_features, "now": new_features}


def performance_budget(runner, params):

    if len(params) == 0:
        params = [None]
    elif len(params) > 1:
        raise ValueError("Expected 1 or 0 items in list %s" % params)

    old_value = runner.budget
    runner.set_performance_cap_budget(params[0] or None)

    return {
        "status": True,
        "output": {"was": old_value}
    }


def ips_whitelist(runner, params):
    """ Replace current list of whitelisted IP networks
    """
    params = params[0]
    runner.set_ips_whitelist(params)
    return {"status": True}


def paths_whitelist(runner, params):
    """ Replace current list of whitelisted paths
    """
    params = params[0]

    runner.set_paths_whitelist(params)
    return {"status": True}


def force_logout(runner, params=None):
    instrumentation_remove(runner)
    runner.logout()
    return {"status": True}


ALL_COMMANDS = {
    "instrumentation_enable": instrumentation_enable,
    "instrumentation_remove": instrumentation_remove,
    "get_bundle": get_bundle,
    "rules_reload": rules_reload,
    "features_get": features_get,
    "features_change": features_change,
    "ips_whitelist": ips_whitelist,
    "paths_whitelist": paths_whitelist,
    "performance_budget": performance_budget,
    "force_logout": force_logout,
    "actions_reload": actions_reload,
    "record_stacktrace": record_stacktrace,
}
COALESCE_COMMANDS = ("actions_reload", "rules_reload",)



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Sqreen attack event helpers and placeholder
"""
import traceback
from logging import getLogger

from ._vendors.sqreen_security_signal_sdk import Signal, Trace
from .config import CONFIG
from .remote_exception import traceback_formatter
from .sanitizer import sanitize, sanitize_attacks, sanitize_exceptions

LOGGER = getLogger(__name__)


def get_context_payload():
    """ Return attack payload dependent on the context, right now stacktrace.
    """
    return {
        "context": {
            "backtrace": list(traceback_formatter(traceback.extract_stack()))
        }
    }


class Attack(object):
    def __init__(self, payload, rule_name):
        self.payload = payload
        self.rule_name = rule_name

    def to_dict(self):
        result = {}
        rule_payload = self.payload.get("rule", {})
        request_payload = self.payload.get("request", {})
        local_payload = self.payload.get("local", {})
        if "name" in rule_payload:
            result["rule_name"] = rule_payload["name"]
        if "rulespack_id" in rule_payload:
            result["rulespack_id"] = rule_payload["rulespack_id"]
        if "test" in rule_payload:
            result["test"] = rule_payload["test"]
        if "infos" in self.payload:
            result["infos"] = self.payload["infos"]
        if "time" in local_payload:
            result["time"] = local_payload["time"]
        if "remote_ip" in request_payload:
            result["client_ip"] = request_payload["remote_ip"]
        if "request" in self.payload:
            result["request"] = self.payload["request"]
        if "params" in self.payload:
            result["params"] = self.payload["params"]
        if "context" in self.payload:
            result["context"] = self.payload["context"]
        if "headers" in self.payload:
            result["headers"] = self.payload["headers"]
        return result


class RequestRecord(object):
    """Request record objects."""

    VERSION = "20171208"

    def __init__(self, payload):
        self.payload = payload

    def to_dict(self):
        """Export the record as a dict object."""
        result = {"version": self.VERSION}
        sensitive_values = None

        if "request" in self.payload:
            request_dict = self.payload["request"]
            result["request"] = request_dict
            if "client_ip" in request_dict:
                result["client_ip"] = request_dict.pop("client_ip")
        else:
            result["request"] = {}
        if "params" in self.payload:
            result["request"]["parameters"] = self.payload["params"]
        if "headers" in self.payload:
            result["request"]["headers"] = self.payload["headers"]

        if CONFIG["STRIP_SENSITIVE_DATA"]:
            sanitized_request, sensitive_values = sanitize(result["request"])
            result["request"] = sanitized_request

        if "response" in self.payload:
            result["response"] = self.payload["response"]

        if "observed" in self.payload:
            observed_dict = self.payload["observed"]
            result["observed"] = observed_dict
            rulespack = None
            attacks = observed_dict.get("attacks", [])
            for attack_dict in attacks:
                rulespack = attack_dict.pop("rulespack_id", None) or rulespack
            if attacks and sensitive_values:
                observed_dict["attacks"] = list(sanitize_attacks(attacks, sensitive_values))
            exceptions = observed_dict.get("sqreen_exceptions", [])
            for exc_dict in exceptions:
                payload_dict = exc_dict.pop("exception", None)
                if payload_dict:
                    exc_dict["message"] = payload_dict["message"]
                    exc_dict["klass"] = payload_dict["klass"]
                rulespack = exc_dict.pop("rulespack_id", None) or rulespack
            if exceptions and sensitive_values:
                observed_dict["sqreen_exceptions"] = list(sanitize_exceptions(exceptions, sensitive_values))
            if rulespack:
                result["rulespack_id"] = rulespack
            if "observations" in observed_dict:
                result["observed"]["observations"] = [
                    {"category": cat, "key": key, "value": value, "time": time}
                    for (cat, time, key, value) in observed_dict[
                        "observations"
                    ]
                ]
            if "sdk" in observed_dict:
                result["observed"]["sdk"] = [
                    {"name": entry[0], "time": entry[1], "args": entry[2:]}
                    for entry in observed_dict["sdk"]
                ]
        if "local" in self.payload:
            result["local"] = self.payload["local"]
        return result

    def to_trace(self):
        """ Request record trace.
        """
        event_payload = self.to_dict()

        context = dict(type="http")
        data = []
        trace = Trace(
            context_schema="http/2020-01-01T00:00:00.000Z",
            context=context,
            actor={},
            data=data,
        )

        rulespack_id = event_payload.get("rulespack_id", "0000000000000000000000000000000000000000")
        for entry in event_payload.get("observed", {}).get("sqreen_exceptions", []):
            signal = Signal(
                type="point",
                signal_name="sq.agent.exception",
                time=entry["time"],
                source="sqreen:rule:{}:{}".format(entry.get("rulespack_id", rulespack_id), entry["rule_name"]),
                payload_schema="exception/2020-01-01T00:00:00.000Z",
                payload={
                    "klass": entry["klass"],
                    "message": entry["message"],
                    "infos": entry.get("infos", {}),
                    "test": entry.get("test", False),
                },
                location={
                    "stack_trace": entry["backtrace"],
                }
            )
            data.append(signal)

        for entry in event_payload.get("observed", {}).get("attacks", []):
            signal = Signal(
                type="point",
                signal_name="sq.agent.attack.{}".format(entry.get("attack_type", "unknown")),
                time=entry["time"],
                source="sqreen:rule:{}:{}".format(entry.get("rulespack_id") or rulespack_id, entry["rule_name"]),
                payload_schema="attack/2020-01-01T00:00:00.000Z",
                payload={
                    "beta": entry.get("beta", False),
                    "test": entry.get("test", False),
                    "block": entry.get("block", False),
                    "infos": entry.get("infos", {}),
                },
            )
            backtrace = entry.get("backtrace")
            if backtrace:
                signal["location"] = {"stack_trace": backtrace}
            data.append(signal)

        user_identifiers = None
        traits = None
        for entry in event_payload.get("observed", {}).get("sdk", []):
            args = entry["args"]
            if entry["name"] == "identify":
                user_identifiers = args[0]
                traits = args[1]
                continue
            elif entry["name"] != "track":
                LOGGER.debug("ignoring sdk event: %r", entry)
                continue
            event_name = args[0]
            if not event_name.startswith("sq."):
                event_name = "sq.sdk.{}".format(event_name)
            options = args[1]
            payload = {"properties": options.get("properties")}
            if user_identifiers is not None:
                payload["user_identifiers"] = options.get("user_identifiers")
            else:
                user_identifiers = options.get("user_identifiers")
            data.append(Signal(
                type="point",
                signal_name=event_name,
                source="sqreen:sdk:track",
                time=options.get("timestamp", entry["time"]),
                payload_schema="track_event/2020-01-01T00:00:00.000Z",
                payload=payload,
            ))

        request = event_payload["request"]
        user_agent = request.get("user_agent")
        if user_agent:
            trace["actor"]["user_agent"] = user_agent

        client_ip = event_payload.get("client_ip")
        if client_ip is not None:
            trace["actor"]["ip_addresses"] = [client_ip]

        if user_identifiers:
            trace["actor"]["user_identifiers"] = user_identifiers
        if traits:
            trace["actor"]["traits"] = traits

        context["request"] = dict(event_payload["request"])
        context["response"] = dict(event_payload.get("response", {}))

        trace["time"] = event_payload.get("local", {}).get("time")

        return trace



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Count http codes for flask framework
"""
import logging

from ..rules import RuleCallback

LOGGER = logging.getLogger(__name__)


class CountHTTPCodesCBFlask(RuleCallback):

    INTERRUPTIBLE = False

    def post(self, instance, args, kwargs, result=None, **options):
        """ Recover the status code and update the http_code metric
        """
        response = result
        status_code = response.status_code
        self.record_observation("http_code", str(status_code), 1)

        return {}

    def failing(self, instance, args, kwargs, **options):
        """ Recover the status code and update the http_code metric
        """
        self.record_observation("http_code", "500", 1)

        return {}



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Signal delivery method
"""
import logging

from ..__about__ import __version__
from .._vendors.sqreen_security_signal_sdk import Client
from ..events import RequestRecord
from ..http_client import USER_AGENT
from ..remote_exception import RemoteException
from .simple import SimpleDeliverer

LOGGER = logging.getLogger(__name__)


class AgentSignalClient(Client):

    user_agent = USER_AGENT


class SignalDeliverer(SimpleDeliverer):

    client_cls = AgentSignalClient

    def __init__(self, session, batch_size, max_staleness):
        super(SignalDeliverer, self).__init__(session)
        self.batch_size = batch_size
        self.original_max_staleness = max_staleness
        self.max_staleness = max_staleness
        self.client = self.client_cls(
            token=session.session_token,
            max_batch_size=self.batch_size,
            interval_batch=self.max_staleness,
            session_token=True,
        )

    def post_event(self, event):
        if isinstance(event, RequestRecord):
            return self.client.trace(**event.to_trace())
        elif isinstance(event, RemoteException):
            return self.client.signal(**event.to_signal())
        LOGGER.debug("Event not supported: %r", event)
        raise NotImplementedError

    def post_metrics(self, metrics):
        for metric in metrics:
            payload = {
                "capture_interval_s": metric["period"],
                "date_started": metric["start"],
                "date_ended": metric["finish"],
            }
            if metric["kind"] == "Binning":
                payload_schema = "metric_binning/2020-01-01T00:00:00.000Z"
                o = metric["observation"]
                payload.update({
                    "unit": o["u"],
                    "base": o["b"],
                    "bins": o["v"],
                    "max": o["v"].pop("max"),
                })
            else:
                payload_schema = "metric/2020-01-01T00:00:00.000Z"
                payload.update({
                    "kind": metric["kind"].lower(),
                    "values": [{"key": str(k), "value": v}
                               for k, v in metric["observation"].items()],
                })

            self.client.metric(
                "sq.agent.metric.{}".format(metric["name"]),
                payload,
                payload_schema=payload_schema,
                source="sqreen:agent:{}".format(__version__),
            )

    def drain(self, resiliently):
        self.client.flush()

    def tick(self):
        self.client.flush(soft=True)



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Record request context
"""

from logging import getLogger

from .record_request import RecordRequest

LOGGER = getLogger(__name__)


class RecordRequestContext(RecordRequest):

    def _clear_request(self):
        if self.runner is None:
            self.storage.clear_request(None, None)
        else:
            self.storage.clear_request(
                self.runner.queue, self.runner.observation_queue
            )

    def post(self, instance, args, kwargs, **options):
        self._clear_request()

    def failing(self, instance, args, kwargs, **options):
        self._clear_request()



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
import sys

from ..._vendors.wrapt import FunctionWrapper
from ...exceptions import RequestBlocked
from ...runtime_storage import runtime
from .base import BaseMiddleware


class FlaskMiddleware(BaseMiddleware):
    def __call__(self, original):

        def wrapper(wrapped, instance, args, kwargs):
            """ Call the lifecycles methods with these arguments:
            Flask pre callbacks will receive these arguments:
            (None)
            Flask post callbacks will receive these arguments:
            (None, response)
            Flask failing callbacks will receive these arguments:
            (None, exception)
            """
            try:
                self.strategy.before_hook_point()

                self.execute_pre_callbacks(record_attack=True)

                try:
                    response = wrapped(*args, **kwargs)
                except Exception:
                    self.execute_failing_callbacks(sys.exc_info())
                    raise
                return self.execute_post_callbacks(
                    response, record_attack=True
                )
            except RequestBlocked:
                runtime.clear_request(self.queue, self.observation_queue)
                raise
            finally:
                runtime.clear_request()

        return FunctionWrapper(original, wrapper)



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
"""Utility functions for asyncio."""

import asyncio
import threading

# Dedicated event loop for calling coroutines in a synchronous context.
_EVENT_LOOP = None


def _run_event_loop(loop):
    """Set the current event loop and run it forever."""
    asyncio.set_event_loop(loop)
    loop.run_forever()


def _get_loop():
    """Return a running event loop.

    If called for the first time, define _EVENT_LOOP and run it forever in a
    dedicated thread.
    """
    global _EVENT_LOOP
    if _EVENT_LOOP is None:
        _EVENT_LOOP = asyncio.new_event_loop()
        thread = threading.Thread(
            target=_run_event_loop, args=(_EVENT_LOOP,), daemon=True
        )
        thread.start()
    return _EVENT_LOOP


def run_coroutine(coro):
    """Run a coroutine in a synchronous context."""
    future = asyncio.run_coroutine_threadsafe(coro=coro, loop=_get_loop())
    return future.result()



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
"""Count HTTP codes for the aiohttp framework."""

import logging

from ..rules import RuleCallback

LOGGER = logging.getLogger(__name__)


class CountHTTPCodesCBAioHTTP(RuleCallback):

    INTERRUPTIBLE = False

    def post(self, instance, args, kwargs, result=None, **options):
        """Recover the status code and update the http_code metric."""
        response = result
        status_code = response.status
        self.record_observation("http_code", str(status_code), 1)
        return {}

    def failing(self, instance, args, kwargs, **options):
        """Recover the status code and update the http_code metric."""
        self.record_observation("http_code", "500", 1)
        return {}



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Look for badly behaved clients
"""
import logging
import sys

from .regexp_rule import RegexpRule

LOGGER = logging.getLogger(__name__)

ERROR_STRING = (
    "<html><head><title>Internal Server Error</title></head>"
    "<body><h1><p>Internal Server Error</p></h1></body></html>"
)


class UserAgentMatchesCB(RegexpRule):
    def pre(self, instance, args, kwargs, **options):
        environ = args[-2]
        start_response = args[-1]

        user_agent = environ.get("HTTP_USER_AGENT")
        if not user_agent:
            return

        match = self.match_regexp(user_agent)
        if not match:
            return

        infos = {"found": match, "in": user_agent}
        self.record_attack(infos)

        # Try to detect gunicorn
        server_is_gunicorn = environ.get("SERVER_SOFTWARE", "").startswith(
            "gunicorn"
        )
        is_gevent_loaded = "gevent" in sys.modules
        if server_is_gunicorn and is_gevent_loaded and self.block:
            # Log the error
            err = "Sqreen blocked a security threat (type: #%s). No action is required."
            LOGGER.error(err, self.rule_name)

            # Return a 500 error instead for gunicorn
            start_response(
                "500 Internal Server Error",
                [
                    ("Content-Type", "text/html"),
                    ("Content-Length", str(len(ERROR_STRING))),
                ],
            )

            return {"status": "override", "new_return_value": [ERROR_STRING]}
        else:
            # Blocking is checked by the caller
            return {
                "status": "raise",
                "data": match,
                "rule_name": self.rule_name,
            }



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#

import logging

from ..actions import ACTION_STORE, ActionName
from ..rules import RuleCallback
from ..sdk import events

LOGGER = logging.getLogger(__name__)


class IPActionCB(RuleCallback):

    INTERRUPTIBLE = False

    def pre(self, instance, args, kwargs, **options):
        request = self.storage.get_current_request()
        if request is None:
            return
        client_ip = request.raw_client_ip
        if client_ip is None:
            return
        action = ACTION_STORE.get_for_ip(client_ip)
        if action is None:
            return
        events.track_action(action, {"ip_address": str(client_ip)},
                            storage=self.storage)
        if action.name == ActionName.BLOCK_IP:
            LOGGER.debug(
                "IP %s is blacklisted by action %s",
                client_ip,
                action.iden,
            )
            return {
                "status": "action_block",
                "action_id": action.iden,
                "immediate": True,
            }
        elif action.name == ActionName.REDIRECT_IP:
            LOGGER.debug(
                "IP %s is redirected to %r by action %s",
                client_ip,
                action.target_url,
                action.iden,
            )
            return {
                "status": "action_redirect",
                "action_id": action.iden,
                "target_url": action.target_url,
                "immediate": True,
            }



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Helpers to retrieve runtime informations
"""
import datetime
import os
import platform
import sys
from hashlib import sha1
from logging import getLogger

import pkg_resources

from .__about__ import __version__
from ._vendors import distro
from .exceptions import UnsupportedFrameworkVersion, UnsupportedPythonVersion
from .webservers import get_webserver

LOGGER = getLogger(__name__)


def get_libsqreen_version():
    """ Return the Sqreen Library version if present.
    """
    try:
        from sq_native import get_lib_version  # type: ignore
    except ImportError:
        return None  # pragma: no cover
    else:
        return get_lib_version()


def get_installed_distributions():
    """ Mimic pip.get_installed_distributions using pkg_resources
    """
    return pkg_resources.working_set


def _format_datetime_isoformat(datetime_object):
    """ Take a datetime and return it as isoformat
    """
    return datetime_object.isoformat()


def get_pid_cmdline(pid):
    """ Try to recover the process parent cmdline
    """
    path = "/proc/{}/cmdline".format(pid)
    try:
        with open(path, "r") as cmdline_file:
            raw_cmdline = cmdline_file.read()

        return raw_cmdline.replace("\x00", " ")
    except Exception:
        LOGGER.warning("Error reading %s path", path)
        return ""


def get_process_cmdline():
    """ Returns the current process cmdline and returns it as a string
    """
    # Retrieve the command used to launch the process
    if hasattr(sys, "argv") and len(sys.argv) >= 1:
        return " ".join(sys.argv)
    else:
        return get_pid_cmdline(os.getpid())


def get_parent_cmdline():
    """ Returns the parent process cmdline and returns it as a string
    """
    return get_pid_cmdline(os.getppid())


def parse_minor_version(version):
    """ Parse a version and return MAJOR, MINOR as a tuple.

    >>> parse_minor_version('1.9.1')
    (1, 9)
    """
    return tuple(map(int, version.split(".")[:2]))


SUPPORTED_PYTHON_VERSIONS = [(2, 7), (3, 4), (3, 5), (3, 6), (3, 7), (3, 8)]


class RuntimeInfos(object):
    """ Helper to collect and return environement informations about the
    runtime
    """

    def all(self):
        """ Returns aggregated infos from the environment
        """
        resultat = {"various_infos": {}}
        resultat.update(self._agent())
        resultat.update(self._framework())
        resultat.update(self._os())
        resultat.update(self._runtime())
        resultat.update(self._bundle_signature())
        resultat["various_infos"].update(self._time())
        resultat["various_infos"].update(self._dependencies())
        resultat["various_infos"].update(self._process())
        resultat["various_infos"].update(self._webserver())
        return resultat

    @staticmethod
    def _webserver():
        """ Returns information about the current webserver
        """
        webserver = get_webserver()
        return {"webserver": webserver.various_infos()}

    @staticmethod
    def _dependencies():
        """ Returns informations about the installed python dependencies
        """
        dep = [
            {"name": package.project_name, "version": package.version}
            for package in get_installed_distributions()
        ]
        return {"dependencies": dep}

    @staticmethod
    def _bundle_signature():
        """ Returns the signature of installed Python dependencies
        """
        pkgvers = sorted(
            (package.project_name, package.version)
            for package in get_installed_distributions()
        )
        pkgvers = "|".join(
            "{}-{}".format(name, version) for name, version in pkgvers
        )
        sig = sha1(pkgvers.encode()).hexdigest()
        return {"bundle_signature": sig}

    @staticmethod
    def _time():
        """ Returns informations about the current time
        """
        return {"time": _format_datetime_isoformat(datetime.datetime.utcnow())}

    @staticmethod
    def _agent():
        """ Returns informations about the agent
        """
        return {
            "agent_type": "python",
            "agent_version": __version__,
            "libsqreen_version": get_libsqreen_version(),
        }

    @staticmethod
    def _get_package_version(package_name):
        for package in get_installed_distributions():
            if package.project_name == package_name:
                return package.version

    def _framework(self):
        """ Returns informations about the web framework.
        Also check version of Django.
        """
        flask = self._get_package_version("Flask")
        django = self._get_package_version("Django")
        pyramid = self._get_package_version("pyramid")
        aiohttp = self._get_package_version("aiohttp")

        if django:
            django_version = parse_minor_version(django)
            # Check for Django version
            if django_version < (1, 6):
                raise UnsupportedFrameworkVersion("django", django)

            return {"framework_type": "Django", "framework_version": django}
        elif flask:
            flask_version = parse_minor_version(flask)
            # Check for Flask version
            if flask_version < (0, 10):
                raise UnsupportedFrameworkVersion("flask", flask)

            return {"framework_type": "Flask", "framework_version": flask}
        elif pyramid:
            pyramid_version = parse_minor_version(pyramid)
            if pyramid_version < (1, 6):
                raise UnsupportedFrameworkVersion("pyramid", pyramid)

            return {"framework_type": "Pyramid", "framework_version": pyramid}
        elif aiohttp:
            aiohttp_version = parse_minor_version(aiohttp)
            if aiohttp_version < (2, 2):
                raise UnsupportedFrameworkVersion("aiohttp", aiohttp)

            return {"framework_type": "aiohttp", "framework_version": aiohttp}
        else:
            return {"framework_type": None, "framework_version": None}

    @staticmethod
    def _os():
        """ Returns informations about the os
        """

        # Compute the OS version
        if sys.platform == "darwin":
            base_os_version = "Mac OS X {}".format(platform.mac_ver()[0])
        elif sys.platform.startswith("linux"):
            base_os_version = "{0[0]} {0[1]}".format(
                distro.linux_distribution()
            )
        else:
            base_os_version = ""

        os_version = "{}/{}".format(base_os_version, "/".join(os.uname()))
        return {
            "os_type": "{}-{}".format(platform.machine(), sys.platform),
            "os_version": os_version,
            "hostname": platform.node(),
        }

    @staticmethod
    def _runtime():
        """ Returns informations about the python runtime
        """

        # Check that the python version is supported
        python_version = tuple(map(int, platform.python_version_tuple()[:2]))
        if python_version not in SUPPORTED_PYTHON_VERSIONS:
            raise UnsupportedPythonVersion(platform.python_version())

        python_build = platform.python_build()
        version = "{} ({}, {})".format(
            platform.python_version(), python_build[0], python_build[1]
        )

        return {
            "runtime_type": platform.python_implementation(),
            "runtime_version": version,
        }

    @staticmethod
    def _process():
        """ Returns informations about the current process
        """
        return {
            "pid": os.getpid(),
            "ppid": os.getppid(),
            "ppid_cmdline": get_parent_cmdline(),
            "euid": os.geteuid(),
            "egid": os.getegid(),
            "uid": os.getuid(),
            "gid": os.getgid(),
            # sys.argv is not always available
            "name": get_process_cmdline(),
        }

    @staticmethod
    def local_infos():
        infos = RuntimeInfos._time()
        infos["name"] = platform.node()
        return infos



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
"""Sqreen Security Signal SDK for Python."""
from .__about__ import __version__
from .client import Client
from .compat_model import Signal, Trace

__all__ = [
    "__version__",
    "Client",
    "Signal",
    "Trace",
]



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Callbacks rules classes and helpers
"""
import logging
from datetime import datetime

from ._vendors.wrapt import FunctionWrapper
from .condition_evaluator import ConditionEvaluator, is_condition_empty
from .constants import LIFECYCLE_METHODS
from .remote_exception import raw_traceback_formatter, traceback_formatter
from .runtime_storage import runtime

LOGGER = logging.getLogger(__name__)


class BaseCallback(object):
    """ Base class for callbacks.

    The hook_name is the path to the hook, it could be either a module, like
    "package.module", or a specific class "package.module::Class".
    The hook_name is the name of the function to hook_on, it's relative to
    the hook_module, for example with a hook_module equal to
    "package.module::Class" and a hook_path equal to "method", we will
    hook on the method named "method" of a class named "Class" in the module
    named "package.module"
    """

    SUPPORTS_BUDGET = False  # override constants in subclasses supporting it
    INTERRUPTIBLE = True

    def __init__(self, hook_module, hook_name, strategy=None):
        self.hook_module = hook_module
        self.hook_name = hook_name
        self.strategy = strategy

    @property
    def collaborative(self):
        """Is the callback collaborative and support running with a budget."""
        return self.SUPPORTS_BUDGET

    @property
    def skippable(self):
        """Allow the callback to be skipped if request is overtime."""
        return self.INTERRUPTIBLE


def check_condition_wrapper(rule_data, function, condition, lifecycle):
    """ Wrapper that will check lifecycle method pre-condition before
    calling it.
    If pre-conditions are true, call the lifecycle method, otherwise
    return None.
    """

    def wrapper(wrapped, callback, args, kwargs):
        """ Wrapper around lifecycle method
        """

        # Get the storage from the callback if available
        storage = callback.storage if callback is not None else runtime

        # Compute return value depending on the lifecycle method we wrap
        return_value = kwargs.get("result") if lifecycle == "post" \
            else kwargs.get("exc_info")

        binding_eval_args = {
            "request": storage.get_current_request(),
            "response": storage.get_current_response(),
            "inst": args[0],
            "args": storage.get_current_args(args[1]),
            "kwargs": args[2],
            "data": rule_data,
            "rv": return_value,
        }

        # Check the pre condition
        condition_result = condition.evaluate(**binding_eval_args)

        LOGGER.debug(
            "ConditionWrapper (%s) result for %s: %s",
            condition,
            function,
            condition_result,
        )

        if condition_result in (False, None):
            return None

        # Execute the hook otherwise with the original args
        return wrapped(*args, **kwargs)

    return FunctionWrapper(function, wrapper)


def call_count_wrapper(function, lifecycle, observation_key):
    """ Wrapper around lifecycle methods that record number of calls
    """

    def wrapper(wrapped, callback, args, kwargs):
        """ Record the number of calls for this callback lifecycle method.
        Buffer the number in the callback itself (self.call_counts) and record
        an observation every X times, X being the field call_count_interval of
        the rule.
        """
        current_count = callback.call_counts[lifecycle]

        if current_count + 1 == callback.call_count_interval:
            callback.record_observation(
                "sqreen_call_counts", observation_key, callback.call_count_interval
            )
            callback.call_counts[lifecycle] = 0
        else:
            callback.call_counts[lifecycle] += 1

        return wrapped(*args, **kwargs)

    return FunctionWrapper(function, wrapper)


def perf_monitoring_wrapper(function, name):
    """ Wrapper that will monitor the performance of the rule callback.
    It will time the execution of the callback and send it as an
    observation metric.

    The wrapper should be called first before every others.
    """

    def wrapper(wrapped, callback, args, kwargs):
        """ Time the callback execution.
        """
        # Get the storage from the callback if available
        storage = callback.storage if callback is not None else runtime

        with storage.trace(name):
            return wrapped(*args, **kwargs)

    return FunctionWrapper(function, wrapper)


class RuleCallback(BaseCallback):

    SECTIONS = ("request", "response", "params", "headers", "context")

    def __init__(
        self,
        hook_module,
        hook_name,
        rule_name,
        rulespack_id,
        runner,
        beta=False,
        block=False,
        test=False,
        strategy_cls=None,
        data=None,
        conditions=None,
        callbacks=None,
        call_count_interval=0,
        metrics=None,
        payload_sections=None,
        priority=None,
        attack_render_info=None,
        attack_type=None,
        storage=runtime,
    ):
        super(RuleCallback, self).__init__(
            hook_module, hook_name, strategy_cls
        )
        self.rule_name = rule_name
        self.rulespack_id = rulespack_id

        if data is None:
            data = {}
        self.data = data
        self.beta = beta
        self.block = block
        self.test = test
        self.runner = runner
        self.metrics = metrics
        self.storage = storage

        if conditions is None:
            conditions = {}
        self.conditions = conditions

        if callbacks is None:
            callbacks = []
        self.callbacks = callbacks
        self.attack_render_info = attack_render_info
        self.attack_type = attack_type

        # Callbacks
        self.call_count_interval = call_count_interval
        self.call_counts = {"pre": 0, "post": 0, "failing": 0}

        self._apply_conditions()
        self._apply_call_count()
        self._apply_perf_monitoring()

        # Payload sections.
        if payload_sections is None:
            payload_sections = self.SECTIONS
        self.payload_sections = payload_sections

        # Priority (lower number means higher priority)
        self.priority = priority if priority is not None else 100

    @classmethod
    def from_rule_dict(cls, rule_dict, runner, storage=runtime):
        """ Return a RuleCallback based on a rule dict
        """
        return cls(
            hook_module=rule_dict["hookpoint"]["klass"],
            hook_name=rule_dict["hookpoint"]["method"],
            rule_name=rule_dict["name"],
            rulespack_id=rule_dict["rulespack_id"],
            strategy_cls=rule_dict["hookpoint"].get("strategy", "import_hook"),
            attack_render_info=rule_dict["hookpoint"].get("attack_render_info", {}),
            beta=rule_dict.get("beta", False),
            block=rule_dict.get("block", False),
            test=rule_dict.get("test", False),
            data=rule_dict.get("data"),
            conditions=rule_dict.get("conditions"),
            callbacks=rule_dict.get("callbacks"),
            call_count_interval=rule_dict.get("call_count_interval", 0),
            runner=runner,
            metrics=rule_dict.get("metrics", []),
            payload_sections=rule_dict.get("payload"),
            priority=rule_dict.get("priority"),
            attack_type=rule_dict.get("attack_type"),
            storage=storage,
        )

    @property
    def performance_monitoring_enabled(self):
        """ Return True if the performance monitoring is enabled for this rule.
        """
        if self.runner is not None:
            return self.runner.performance_metrics and self.runner.performance_metrics_settings.enabled()
        return False

    def get_remaining_budget(self, callback_kwargs=None):
        """ Get the remaining time we have for running this callback according
        to the request budget if there is one. The budget can be overriden
        with a special option in kwargs. Return None otherwise.
        """
        if callback_kwargs is not None:
            override_budget = callback_kwargs.pop("__sqreen_override_budget", None)
            if override_budget is not None:
                return override_budget
        if self.runner is not None and self.performance_monitoring_enabled:
            budget = self.runner.budget
            if budget is not None:
                return budget - (self.storage.request_sq_time or 0)

    def _apply_conditions(self):
        """ Wrap each lifecycle methods if the Rule define them and if we have
        conditions for them.
        """
        for lifecycle in LIFECYCLE_METHODS.values():
            conditions = self.conditions.get(lifecycle)

            if not is_condition_empty(conditions) and hasattr(self, lifecycle):
                conditions = ConditionEvaluator(conditions)

                # Wrap the lifecycle method
                lifecycle_method = getattr(self, lifecycle)
                wrapped = check_condition_wrapper(
                    self.data, lifecycle_method, conditions, lifecycle)
                setattr(self, lifecycle, wrapped)

    def _apply_call_count(self):
        # Only count calls if call_count_interval is > 0
        if self.call_count_interval == 0:
            return

        for lifecycle in LIFECYCLE_METHODS.values():
            if hasattr(self, lifecycle):
                lifecycle_method = getattr(self, lifecycle)

                observation_key = "%s/%s/%s" % (
                    self.rulespack_id,
                    self.rule_name,
                    lifecycle,
                )

                wrapped = call_count_wrapper(
                    lifecycle_method, lifecycle, observation_key)
                setattr(self, lifecycle, wrapped)

    def _apply_perf_monitoring(self):
        # Only enable the performance monitoring if the feature flag is
        if not self.performance_monitoring_enabled:
            return

        for lifecycle in LIFECYCLE_METHODS.values():
            if hasattr(self, lifecycle):
                lifecycle_method = getattr(self, lifecycle)

                metric_name = "sq.{}.{}".format(self.rule_name, lifecycle)

                # Register the performance metric if it does not exist yet
                if self.runner is not None \
                        and not self.runner.metrics_store.is_registered(metric_name):
                    self.runner.metrics_store.register_metric(
                        metric_name, "Binning",
                        self.runner.performance_metrics_settings.period,
                        base=self.runner.performance_metrics_settings.base,
                        unit=self.runner.performance_metrics_settings.unit
                    )

                wrapped = perf_monitoring_wrapper(
                    lifecycle_method, metric_name)
                setattr(self, lifecycle, wrapped)

    def exception_infos(self, infos={}):
        """ Returns additional infos in case of exception
        """
        return {"rule_name": self.rule_name, "rulespack_id": self.rulespack_id}

    def record_attack(self, infos=None, at=None):
        """Record an attack."""
        if at is None:
            at = datetime.utcnow()
        payload = {
            "infos": infos,
            "rulespack_id": self.rulespack_id,
            "rule_name": self.rule_name,
            "beta": self.beta,
            "block": self.block,
            "test": self.test,
            "attack_type": self.attack_type,
            "time": at,
        }
        if "context" in self.payload_sections:
            current_request = self.storage.get_current_request()
            if current_request:
                payload["backtrace"] = list(current_request.raw_caller)
        LOGGER.debug("Observed attack %r", payload)
        self.storage.observe("attacks", payload, self.payload_sections)

    def record_observation(self, metric_name, key, value, at=None):
        """Record a metric observation."""

        self.global_record_observation(metric_name, key, value, at, storage=self.storage)

    @classmethod
    def global_record_observation(cls, metric_name, key, value, at=None, storage=runtime):
        """Record a metric observation"""

        if at is None:
            at = datetime.utcnow()
        payload = (metric_name, at, key, value)
        LOGGER.debug("Observed metric %r", payload)
        storage.observe("observations", payload, report=False)

    def record_exception(self, exc, exc_info, stack=None, infos=None, at=None):
        """Record an exception."""
        if infos is None:
            infos = {}
        if at is None:
            at = datetime.utcnow()
        # Try to recover some infos from the exception if it's a
        # SqreenException.
        try:
            infos.update(exc.exception_infos())
        except Exception:
            pass
        bt = raw_traceback_formatter(exc_info[2])
        if stack is not None:
            bt = traceback_formatter(stack) + bt
        payload = {
            "message": str(exc_info[1]),
            "klass": exc_info[0].__name__,
            "infos": infos,
            "rule_name": self.rule_name,
            "rulespack_id": self.rulespack_id,
            "test": self.test,
            "time": at,
            "backtrace": bt,
        }
        LOGGER.debug("Observed exception %r", payload)
        self.storage.observe("sqreen_exceptions", payload, self.payload_sections)

    def record_overtime(self, lifecycle=None, at=None):
        """Record an overtime execution."""
        if at is None:
            at = datetime.utcnow()
        name = self.rule_name
        if lifecycle is not None:
            name += ".{}".format(lifecycle)
        payload = ("request_overbudget_cb", at, name, 1)
        LOGGER.debug("Observed overbudget callback %r", payload)
        self.storage.observe("overtime_callbacks", payload, report=False)

    @property
    def lifecycle_methods(self):
        return [
            lifecycle
            for lifecycle in LIFECYCLE_METHODS.values()
            if hasattr(self, lifecycle)
        ]

    @property
    def whitelisted(self):
        """ Return True if the callback should be skipped (whitelisted
        request), False otherwise
        """
        return self.storage.get_whitelist_match() or False

    def __repr__(self):
        return "%s(rule_name=%r)" % (self.__class__.__name__, self.rule_name)



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
import sys

from ..._vendors.wrapt import FunctionWrapper
from ...exceptions import RequestBlocked
from ...frameworks.pyramid_framework import PyramidResponse
from ...runtime_storage import runtime
from .base import BaseMiddleware


class PyramidMiddleware(BaseMiddleware):

    def __call__(self, original):

        def wrapper(wrapped, instance, args, kwargs):
            """ Call the lifecycles methods with these arguments:
            Pyramid pre callbacks will receive these arguments:
            (None, request)
            Pyramid post callbacks will receive these arguments:
            (None, response)
            Pyramid failing callbacks will receive these arguments:
            (None, exception)
            """
            from pyramid.response import Response  # type: ignore

            try:
                request = args[0]
                self.strategy.before_hook_point()
                pre_args = (request,)
                self.execute_pre_callbacks(pre_args, record_attack=True)

                try:
                    response = wrapped(*args, **kwargs)
                except Exception as e:
                    if isinstance(e, Response):
                        self.execute_post_callbacks(e)
                    else:
                        self.execute_failing_callbacks(sys.exc_info())
                    raise

                return self.execute_post_callbacks(
                    response, record_attack=True
                )
            except RequestBlocked:
                is_uwsgi = "uwsgi.version" in request.environ

                # In uwsgi 2.0.14, raising an AttackBlocked kill the connection
                # Instead return a 500 error
                if is_uwsgi:
                    response = Response("Internal Server Error")
                    response.status_int = 500
                    runtime.store_response(PyramidResponse(response))
                    runtime.clear_request(self.queue, self.observation_queue)
                    return response

                runtime.clear_request(self.queue, self.observation_queue)
                raise
            finally:
                runtime.clear_request()

        return FunctionWrapper(original, wrapper)



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
import sys

if sys.version_info >= (3, 8):
    # The syntax of this module is not compatible with Python 2.7, lazy loading it.

    from .model import AnySignal, Batch, SignalType, Signal, Trace

elif sys.version_info >= (3, 5):
    # Fallback to more basic types.

    from enum import Enum
    from typing import Any, Dict, List, Union

    class SignalType(Enum):
        POINT = "point"
        METRIC = "metric"

    class Signal(Dict[str, Any]):
        pass

    class Trace(Dict[str, Any]):
        pass

    AnySignal = Union[Signal, Trace]

    class Batch(List[AnySignal]):
        pass

else:
    # Compatibility types for Python environments without the typing module.
    class SignalType:
        class POINT:
            name = "POINT"
            value = "point"

        class METRIC:
            name = "METRIC"
            value = "metric"

    class Signal(dict):
        """
        Compatibility type for signals.
        """

    class Trace(dict):
        """
        Compatibility type for traces.
        """

    class Batch(list):
        """
        Compatibility type for batches.
        """



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Instrumentation helper responsible for adding dynamic callback
"""
import logging

from ..remote_exception import RemoteException
from ..utils import HAS_ASYNCIO
from .import_hook import ModuleFinder
from .strategies import (
    DBApi2Strategy,
    DjangoStrategy,
    FlaskStrategy,
    ImportHookStrategy,
    LXMLResolverStrategy,
    Psycopg2Strategy,
    PyramidStrategy,
    WSGIFactoryStrategy,
    WSGIReceiverStrategy,
    WSGIStrategy,
)

STRATEGIES = {
    "import_hook": ImportHookStrategy,
    "DBApi2": DBApi2Strategy,
    "django": DjangoStrategy,
    "psycopg2": Psycopg2Strategy,
    "flask": FlaskStrategy,
    "pyramid": PyramidStrategy,
    "wsgi": WSGIStrategy,
    "wsgi_factory": WSGIFactoryStrategy,
    "wsgi_receiver": WSGIReceiverStrategy,
    "lxml_resolver": LXMLResolverStrategy,
}

if HAS_ASYNCIO:
    from .strategies import (
        AioHTTPHookStrategy,
        AioHTTPInstallStrategy,
        AsyncEventLoopStrategy,
        AsyncImportHookStrategy,
    )

    STRATEGIES.update({
        "async_event_loop": AsyncEventLoopStrategy,
        "async_import_hook": AsyncImportHookStrategy,
        "aiohttp_install": AioHTTPInstallStrategy,
        "aiohttp_hook": AioHTTPHookStrategy,
    })

LOGGER = logging.getLogger(__name__)


class Instrumentation(object):
    """ The instrumentation class is the exposed face of the
    instrumentation engine. It dispatchs to the right strategy,
    the default one is set_attr.

    The instrumentation class takes attack queue and observation queue as parameters.

    The instrument class dispatch to different strategies based on strategy
    name defined in callback. It ask stategy for an unique id based on hook path
    infos and ensure to have only one strategy instance per id. It's needed
    for DBApi2 strategy where every sqlite3 callbacks will be stored in the same
    strategy to avoid double-instrumentation.

    It also instantiate a global ImportHook that strategy will register
    against when hooking.

    TODO:
        - Store callbacks in a single callback tree
        - Transform the remaining strategies into stateless patchers
    """

    def __init__(self, observation_queue, queue, before_hook_point=None):
        self.observation_queue = observation_queue
        self.queue = queue
        self.strategies = {}
        self.before_hook_point = before_hook_point
        self.import_hook = ModuleFinder()
        self.enabled = False

    def add_callback(self, callback):
        """ Add a callback. The callback defines itself where it should
        hook to and the strategy use for hooking (set_attr or DBApi2)
        """
        try:
            strategy_class = self._get_strategy_class(callback.strategy)

            # Get the strategy id
            strategy_id = strategy_class.get_strategy_id(callback)

            # Check if we already have a strategy
            if strategy_id in self.strategies:
                strategy_instance = self.strategies[strategy_id]
                LOGGER.debug(
                    "Reusing strategy %s for id %s",
                    strategy_instance,
                    strategy_id,
                )
            else:
                strategy_instance = strategy_class(
                    strategy_id,
                    self.observation_queue,
                    self.queue,
                    self.import_hook,
                    self.before_hook_point,
                )
                LOGGER.debug(
                    "Instantiate strategy %s for id %s -> %s",
                    strategy_class,
                    strategy_id,
                    strategy_instance,
                )
                self.strategies[strategy_id] = strategy_instance

            strategy_instance.add_callback(callback)
        except Exception:
            # If the strategy fails to hook either at instantiation
            # or hook method, catch the exception and log it
            LOGGER.exception(
                "Callback %r fails to hook", callback, exc_info=True
            )

            # And send the exception back to the backend
            infos = {"callback": callback.__dict__}
            remote_exception = RemoteException.from_exc_info(callback_payload=infos)
            self.queue.put(remote_exception)

    def deinstrument(self, callback):
        """ Deactive instrumentation on the callback endpoint
        """
        strategy_class = self._get_strategy_class(callback.strategy)

        # Get the strategy id
        strategy_id = strategy_class.get_strategy_id(callback)

        self.strategies[strategy_id].deinstrument(callback)

    def deinstrument_all(self):
        """ Deactive instrumentation on all callbacks by calling
        deinstrument_all on all strategies
        """
        for strategy in self.strategies.values():
            strategy.deinstrument_all()
        self.enabled = False

    def hook_all(self):
        """ First, inject the import hook in sys.meta if not present.
        Then hook all strategies, must be called after all the callbacks has
        been added. The call to strategies.hook will register patcher
        to the actual hook.
        Then call import_hook.apply_patchers that will apply patcher on
        modules already imported.
        """
        self.import_hook.inject()
        for strategy in self.strategies.values():
            strategy.hook()
        self.import_hook.apply_patchers()
        self.enabled = True

    @staticmethod
    def _get_strategy_class(strategy):
        """ Return a strategy class depending on the strategy name passed
        in parameter.
        Raise a NotImplementedError if the strategy is unknown.
        """
        strategy_cls = STRATEGIES.get(strategy)
        if strategy_cls is None:
            raise NotImplementedError(
                "Unknown hooking strategy {!r}".format(strategy))
        return strategy_cls



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Generic WSGI HTTP Request / Response stuff
"""
import sys
from collections import defaultdict
from itertools import chain
from logging import getLogger

from ..exceptions import MissingDataException
from ..utils import to_unicode_safe
from .base import BaseRequest
from .ip_utils import get_real_user_ip

if sys.version_info[0] < 3:
    import urlparse
else:
    import urllib.parse as urlparse


LOGGER = getLogger(__name__)


class PHPRequest(BaseRequest):
    """ Helper around PHP request
    """

    def __init__(self, environ=None, storage=None):
        super(PHPRequest, self).__init__(storage=storage)
        if environ is None:
            environ = {}
        self.environ = environ.get("request", {})

    @property
    def body(self):
        """Return the body of the request if present"""
        return self.environ.get("BODY")

    @property
    def query_params(self):
        """ Return parsed query string from request
        """
        return self.environ.get("PARSED_REQ_PARAMS", {}).get('"GET"', {})

    @property
    def form_params(self):
        return self.environ.get("PARSED_REQ_PARAMS", {}).get('"POST"', {})

    @property
    def cookies_params(self):
        return self.environ.get("PARSED_REQ_PARAMS", {}).get('"COOKIE"', {})

    @property
    def query_params_values(self):
        """ Return only query values as a list
        """
        return list(chain.from_iterable(self.query_params.values()))

    @property
    def remote_addr(self):
        """Remote IP address."""
        return self.environ.get("REMOTE_ADDR")

    @property
    def raw_client_ip(self):
        return get_real_user_ip(
            self.environ.get("REMOTE_ADDR"), *self.iter_client_ips()
        )

    @property
    def hostname(self):
        return to_unicode_safe(self.get_raw_header("HTTP_HOST", self.environ.get("SERVER_NAME")))

    @property
    def method(self):
        return self.environ.get("REQUEST_METHOD", "")

    @property
    def client_user_agent(self):
        return to_unicode_safe(self.get_raw_header("HTTP_USER_AGENT", ""))

    @property
    def referer(self):
        return to_unicode_safe(self.get_raw_header("HTTP_REFERER", None))

    @property
    def scheme(self):
        return self.environ.get("URL_SCHEME")

    @property
    def server_port(self):
        return self.environ.get("SERVER_PORT")

    @property
    def remote_port(self):
        return self.environ.get("REMOTE_PORT")

    @property
    def path(self):
        request_uri = self.environ.get("REQUEST_URI", "")
        try:
            # REQUEST_URI can sometimes contain a full URL
            # in this case we parse it and try to extract the path.
            return urlparse.urlparse(request_uri).path
        except ValueError:
            # If the parsing has failed, warn about it and
            # return the initial value. In fact, some path
            # can be misinterpreted as invalid URL by python's
            # urlparse (for example //test] is an invalid IPv6
            # URL but a somewhat valid path).
            LOGGER.debug("request with invalid URI path: %s",
                         request_uri, exc_info=True)
        return request_uri

    @property
    def raw_headers(self):
        return self.environ.get("HEADERS", {})

    @property
    def request_headers(self):
        return defaultdict(lambda: None, **self.raw_headers)

    @property
    def view_params(self):
        return {}

    @property
    def json_params(self):
        return {}

    @property
    def caller(self):
        return [
            "#{count} {file_}({line}) {class_}{type_}{function}".format(
                count=i,
                file_=line.get("file", ""),
                line=line.get("line", "0"),
                class_=line.get("class", ""),
                type_=line.get("type", ""),
                function=line.get("function", ""),
            )
            for i, line in enumerate(self.raw_caller)
        ]

    @property
    def raw_caller(self):
        """The caller is only known from the PHP extension but not sent by default
        to the daemon. We then keep track of the data we are missing and notify the
        daemon by raising a MissingDataException.

        The extension data is not kept inside the request because if the rule is executed
        outside a request context, the request instance is a placeholder and won't
        survive the callback execution."""
        from ..runtime_storage import runtime

        storage = self.storage or runtime
        value = storage.get_extension_data("#.caller")
        if value is None:
            raise MissingDataException("#.caller")
        return value



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
from .engine import Instrumentation

__all__ = ["Instrumentation"]



# -*- coding: utf-8 -*-
#
#  Copyright 2011 Sybren A. Stüvel <sybren@stuvel.eu>
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Commandline scripts.

These scripts are called by the executables defined in setup.py.
"""

from __future__ import with_statement, print_function

import abc
import sys
from optparse import OptionParser

from . import bigfile as rsa_bigfile
from . import key as rsa_key
from . import pkcs1 as rsa_pkcs1

HASH_METHODS = sorted(rsa_pkcs1.HASH_METHODS.keys())


def keygen():
    """Key generator."""

    # Parse the CLI options
    parser = OptionParser(usage='usage: %prog [options] keysize',
                          description='Generates a new RSA keypair of "keysize" bits.')

    parser.add_option('--pubout', type='string',
                      help='Output filename for the public key. The public key is '
                           'not saved if this option is not present. You can use '
                           'pyrsa-priv2pub to create the public key file later.')

    parser.add_option('-o', '--out', type='string',
                      help='Output filename for the private key. The key is '
                           'written to stdout if this option is not present.')

    parser.add_option('--form',
                      help='key format of the private and public keys - default PEM',
                      choices=('PEM', 'DER'), default='PEM')

    (cli, cli_args) = parser.parse_args(sys.argv[1:])

    if len(cli_args) != 1:
        parser.print_help()
        raise SystemExit(1)

    try:
        keysize = int(cli_args[0])
    except ValueError:
        parser.print_help()
        print('Not a valid number: %s' % cli_args[0], file=sys.stderr)
        raise SystemExit(1)

    print('Generating %i-bit key' % keysize, file=sys.stderr)
    (pub_key, priv_key) = rsa_key.newkeys(keysize)

    # Save public key
    if cli.pubout:
        print('Writing public key to %s' % cli.pubout, file=sys.stderr)
        data = pub_key.save_pkcs1(format=cli.form)
        with open(cli.pubout, 'wb') as outfile:
            outfile.write(data)

    # Save private key
    data = priv_key.save_pkcs1(format=cli.form)

    if cli.out:
        print('Writing private key to %s' % cli.out, file=sys.stderr)
        with open(cli.out, 'wb') as outfile:
            outfile.write(data)
    else:
        print('Writing private key to stdout', file=sys.stderr)
        sys.stdout.write(data)


class CryptoOperation(object):
    """CLI callable that operates with input, output, and a key."""

    __metaclass__ = abc.ABCMeta

    keyname = 'public'  # or 'private'
    usage = 'usage: %%prog [options] %(keyname)s_key'
    description = None
    operation = 'decrypt'
    operation_past = 'decrypted'
    operation_progressive = 'decrypting'
    input_help = 'Name of the file to %(operation)s. Reads from stdin if ' \
                 'not specified.'
    output_help = 'Name of the file to write the %(operation_past)s file ' \
                  'to. Written to stdout if this option is not present.'
    expected_cli_args = 1
    has_output = True

    key_class = rsa_key.PublicKey

    def __init__(self):
        self.usage = self.usage % self.__class__.__dict__
        self.input_help = self.input_help % self.__class__.__dict__
        self.output_help = self.output_help % self.__class__.__dict__

    @abc.abstractmethod
    def perform_operation(self, indata, key, cli_args=None):
        """Performs the program's operation.

        Implement in a subclass.

        :returns: the data to write to the output.
        """

    def __call__(self):
        """Runs the program."""

        (cli, cli_args) = self.parse_cli()

        key = self.read_key(cli_args[0], cli.keyform)

        indata = self.read_infile(cli.input)

        print(self.operation_progressive.title(), file=sys.stderr)
        outdata = self.perform_operation(indata, key, cli_args)

        if self.has_output:
            self.write_outfile(outdata, cli.output)

    def parse_cli(self):
        """Parse the CLI options

        :returns: (cli_opts, cli_args)
        """

        parser = OptionParser(usage=self.usage, description=self.description)

        parser.add_option('-i', '--input', type='string', help=self.input_help)

        if self.has_output:
            parser.add_option('-o', '--output', type='string', help=self.output_help)

        parser.add_option('--keyform',
                          help='Key format of the %s key - default PEM' % self.keyname,
                          choices=('PEM', 'DER'), default='PEM')

        (cli, cli_args) = parser.parse_args(sys.argv[1:])

        if len(cli_args) != self.expected_cli_args:
            parser.print_help()
            raise SystemExit(1)

        return cli, cli_args

    def read_key(self, filename, keyform):
        """Reads a public or private key."""

        print('Reading %s key from %s' % (self.keyname, filename), file=sys.stderr)
        with open(filename, 'rb') as keyfile:
            keydata = keyfile.read()

        return self.key_class.load_pkcs1(keydata, keyform)

    def read_infile(self, inname):
        """Read the input file"""

        if inname:
            print('Reading input from %s' % inname, file=sys.stderr)
            with open(inname, 'rb') as infile:
                return infile.read()

        print('Reading input from stdin', file=sys.stderr)
        return sys.stdin.read()

    def write_outfile(self, outdata, outname):
        """Write the output file"""

        if outname:
            print('Writing output to %s' % outname, file=sys.stderr)
            with open(outname, 'wb') as outfile:
                outfile.write(outdata)
        else:
            print('Writing output to stdout', file=sys.stderr)
            sys.stdout.write(outdata)


class EncryptOperation(CryptoOperation):
    """Encrypts a file."""

    keyname = 'public'
    description = ('Encrypts a file. The file must be shorter than the key '
                   'length in order to be encrypted. For larger files, use the '
                   'pyrsa-encrypt-bigfile command.')
    operation = 'encrypt'
    operation_past = 'encrypted'
    operation_progressive = 'encrypting'

    def perform_operation(self, indata, pub_key, cli_args=None):
        """Encrypts files."""

        return rsa_pkcs1.encrypt(indata, pub_key)


class DecryptOperation(CryptoOperation):
    """Decrypts a file."""

    keyname = 'private'
    description = ('Decrypts a file. The original file must be shorter than '
                   'the key length in order to have been encrypted. For larger '
                   'files, use the pyrsa-decrypt-bigfile command.')
    operation = 'decrypt'
    operation_past = 'decrypted'
    operation_progressive = 'decrypting'
    key_class = rsa_key.PrivateKey

    def perform_operation(self, indata, priv_key, cli_args=None):
        """Decrypts files."""

        return rsa_pkcs1.decrypt(indata, priv_key)


class SignOperation(CryptoOperation):
    """Signs a file."""

    keyname = 'private'
    usage = 'usage: %%prog [options] private_key hash_method'
    description = ('Signs a file, outputs the signature. Choose the hash '
                   'method from %s' % ', '.join(HASH_METHODS))
    operation = 'sign'
    operation_past = 'signature'
    operation_progressive = 'Signing'
    key_class = rsa_key.PrivateKey
    expected_cli_args = 2

    output_help = ('Name of the file to write the signature to. Written '
                   'to stdout if this option is not present.')

    def perform_operation(self, indata, priv_key, cli_args):
        """Signs files."""

        hash_method = cli_args[1]
        if hash_method not in HASH_METHODS:
            raise SystemExit('Invalid hash method, choose one of %s' %
                             ', '.join(HASH_METHODS))

        return rsa_key.sign(indata, priv_key, hash_method)


class VerifyOperation(CryptoOperation):
    """Verify a signature."""

    keyname = 'public'
    usage = 'usage: %%prog [options] public_key signature_file'
    description = ('Verifies a signature, exits with status 0 upon success, '
                   'prints an error message and exits with status 1 upon error.')
    operation = 'verify'
    operation_past = 'verified'
    operation_progressive = 'Verifying'
    key_class = rsa_key.PublicKey
    expected_cli_args = 2
    has_output = False

    def perform_operation(self, indata, pub_key, cli_args):
        """Verifies files."""

        signature_file = cli_args[1]

        with open(signature_file, 'rb') as sigfile:
            signature = sigfile.read()

        try:
            rsa_pkcs1.verify(indata, signature, pub_key)
        except rsa_pkcs1.VerificationError:
            raise SystemExit('Verification failed.')

        print('Verification OK', file=sys.stderr)


class BigfileOperation(CryptoOperation):
    """CryptoOperation that doesn't read the entire file into memory."""

    def __init__(self):
        CryptoOperation.__init__(self)

        self.file_objects = []

    def __del__(self):
        """Closes any open file handles."""

        for fobj in self.file_objects:
            fobj.close()

    def __call__(self):
        """Runs the program."""

        (cli, cli_args) = self.parse_cli()

        key = self.read_key(cli_args[0], cli.keyform)

        # Get the file handles
        infile = self.get_infile(cli.input)
        outfile = self.get_outfile(cli.output)

        # Call the operation
        print(self.operation_progressive.title(), file=sys.stderr)
        self.perform_operation(infile, outfile, key, cli_args)

    def get_infile(self, inname):
        """Returns the input file object"""

        if inname:
            print('Reading input from %s' % inname, file=sys.stderr)
            fobj = open(inname, 'rb')
            self.file_objects.append(fobj)
        else:
            print('Reading input from stdin', file=sys.stderr)
            fobj = sys.stdin

        return fobj

    def get_outfile(self, outname):
        """Returns the output file object"""

        if outname:
            print('Will write output to %s' % outname, file=sys.stderr)
            fobj = open(outname, 'wb')
            self.file_objects.append(fobj)
        else:
            print('Will write output to stdout', file=sys.stderr)
            fobj = sys.stdout

        return fobj


class EncryptBigfileOperation(BigfileOperation):
    """Encrypts a file to VARBLOCK format."""

    keyname = 'public'
    description = ('Encrypts a file to an encrypted VARBLOCK file. The file '
                   'can be larger than the key length, but the output file is only '
                   'compatible with Python-RSA.')
    operation = 'encrypt'
    operation_past = 'encrypted'
    operation_progressive = 'encrypting'

    def perform_operation(self, infile, outfile, pub_key, cli_args=None):
        """Encrypts files to VARBLOCK."""

        return rsa_bigfile.encrypt_bigfile(infile, outfile, pub_key)


class DecryptBigfileOperation(BigfileOperation):
    """Decrypts a file in VARBLOCK format."""

    keyname = 'private'
    description = ('Decrypts an encrypted VARBLOCK file that was encrypted '
                   'with pyrsa-encrypt-bigfile')
    operation = 'decrypt'
    operation_past = 'decrypted'
    operation_progressive = 'decrypting'
    key_class = rsa_key.PrivateKey

    def perform_operation(self, infile, outfile, priv_key, cli_args=None):
        """Decrypts a VARBLOCK file."""

        return rsa_bigfile.decrypt_bigfile(infile, outfile, priv_key)


encrypt = EncryptOperation()
decrypt = DecryptOperation()
sign = SignOperation()
verify = VerifyOperation()
encrypt_bigfile = EncryptBigfileOperation()
decrypt_bigfile = DecryptBigfileOperation()



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Utils functions to retrieve the client user ip
"""

from ..utils import ip_address


def get_real_user_ip(remote_addr, *ips):
    """ Try to compute the real user ip from various headers
    """
    private_ip = None
    for list_ips in ips:
        for ip in list_ips.split(","):
            ip = ip.strip()
            if not ip:
                # Dummy IP address, skip it.
                continue
            try:
                # Azure append the port to IPv4 in XFF headers. We need to strip it
                if '.' in ip and ':' in ip:
                    ip = ip.split(':')[0]

                ip = ip_address(ip)
            except ValueError:
                continue
            if ip.is_global:
                return ip
            elif private_ip is None and not ip.is_loopback and ip.is_private:
                private_ip = ip
    # If no global IP was found so far.
    try:
        remote_ip = ip_address(remote_addr)
    except ValueError:
        return private_ip  # May be None.
    if private_ip is None or remote_ip.is_global:
        return remote_ip
    else:
        return private_ip



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
"""Actions for security responses."""

import logging
from collections import defaultdict
from time import time

from .ip_radix import Radix

LOGGER = logging.getLogger(__name__)


class ActionName(object):
    """Enumeration of action names."""

    BLOCK_IP = "block_ip"
    REDIRECT_IP = "redirect_ip"
    BLOCK_USER = "block_user"
    REDIRECT_USER = "redirect_user"


_AVAILABLE_ACTIONS = {}


def register_action(name):
    """Decorator function to register an action."""

    def decorator(action_cls):
        _AVAILABLE_ACTIONS[name] = action_cls
        action_cls.name = name
        return action_cls

    return decorator


class BaseAction(object):
    """Base class for actions."""

    name = None

    def __init__(self, iden, params, duration=None, send_response=True):
        self.iden = iden
        self.params = params
        self.duration = duration
        if duration is not None:
            self.timeout = time() + duration
        else:
            self.timeout = None
        self.send_response = send_response

    def __repr__(self):
        return "{}({!r}, {!r}, {!r}, {!r})".format(
            self.__class__.__name__,
            self.iden,
            self.params,
            self.duration,
            self.send_response,
        )

    def to_dict(self):
        """Convert the action into a dict."""
        return {
            "action": self.name,
            "action_id": self.iden,
            "parameters": self.params,
            "duration": self.duration,
            "send_response": self.send_response,
        }


class BlockIPActionContainer:
    """Deny access based on IP blacklist."""

    def __init__(self, action):
        self.ip_networks = Radix(None)
        self.insert(action)

    def insert(self, action):
        """Insert a new IP"""
        for blacklist in action.params["ip_cidr"]:
            self.ip_networks.insert(blacklist, '/' in blacklist, action)

    def remove(self, action):
        """Remove the IP from an action"""
        for blacklist in action.params["ip_cidr"]:
            self.ip_networks.remove(blacklist, '/' in blacklist)

    def actions(self):
        return self.ip_networks.all_payloads()


@register_action(ActionName.BLOCK_IP)
class BlockIPAction(BaseAction):
    """Action containing an IP block"""
    def __init__(self, *args, **kwargs):
        super(BlockIPAction, self).__init__(*args, **kwargs)
        self.name = ActionName.BLOCK_IP


@register_action(ActionName.REDIRECT_IP)
class RedirectIPAction(BaseAction):
    """Action containing a redirection on IP match"""
    def __init__(self, *args, **kwargs):
        super(RedirectIPAction, self).__init__(*args, **kwargs)
        self.target_url = self.params["url"]
        self.name = ActionName.REDIRECT_IP


@register_action(ActionName.BLOCK_USER)
class BlockUserAction(BaseAction):
    """Block an authenticated user."""

    def __init__(self, *args, **kwargs):
        super(BlockUserAction, self).__init__(*args, **kwargs)
        self.name = ActionName.BLOCK_USER
        self.users = self.params["users"]


@register_action(ActionName.REDIRECT_USER)
class RedirectUserAction(BaseAction):
    """Redirects an authenticated user."""

    def __init__(self, *args, **kwargs):
        super(RedirectUserAction, self).__init__(*args, **kwargs)
        self.name = ActionName.REDIRECT_USER
        self.users = self.params["users"]
        self.target_url = self.params["url"]


class UnsupportedAction(Exception):
    """Exception raised when an action is not supported."""

    def __init__(self, action_name):
        self.action_name = action_name


def action_from_dict(data):
    """Load an action from a dict."""
    action_name = data["action"]
    if action_name not in _AVAILABLE_ACTIONS:
        raise UnsupportedAction(action_name)
    action_cls = _AVAILABLE_ACTIONS[action_name]
    action = action_cls(
        iden=data.get("action_id"),
        params=data.get("parameters"),
        duration=data.get("duration"),
        send_response=data.get("send_response", True),
    )
    return action


class ActionStore:
    """A store to manage all actions."""

    def __init__(self):
        self._actions = defaultdict(list)

    def clear(self):
        """Remove all actions from the store."""
        self._actions.clear()

    def add(self, action):
        """Add an action to the store."""

        # If we're blocking an IP, we will only use a single action in order to share the Radix tree
        if action.name in {ActionName.BLOCK_IP, ActionName.REDIRECT_IP}:
            if self._actions[action.name]:
                self._actions[action.name][0].insert(action)
            else:
                self._actions[action.name].append(BlockIPActionContainer(action))
        else:
            self._actions[action.name].append(action)

    # used by the daemon
    def all_actions(self):
        for action_type, act_list in self._actions.items():
            if (action_type in {ActionName.BLOCK_IP, ActionName.REDIRECT_IP}
                    and len(act_list) == 1
                    and isinstance(act_list[0], BlockIPActionContainer)):
                for a in act_list[0].actions():
                    yield a
            else:
                # the actions are simply stored in a list o/wise
                for a in act_list:
                    yield a

    def _delete_expired(self, action_name, now):
        """Delete expired actions with name *action_name*."""
        self._actions[action_name] = [
            action
            for action in self._actions[action_name]
            if not action.timeout or action.timeout >= now
        ]

    def _get_for_ip(self, action_name, ip):
        """Return the action matching an IP address, or None."""
        action_list = self._actions[action_name]
        if not action_list:
            return None

        # Do we match something?
        match = action_list[0].ip_networks.match(ip)
        if match is not None:
            # Expired action?
            if match.timeout is None or match.timeout >= time():
                return match
            action_list[0].remove(match)
        return None

    def get_for_ip(self, ip):
        """Return the action matching an IP address, or None."""
        return self._get_for_ip(ActionName.BLOCK_IP, ip) or self._get_for_ip(ActionName.REDIRECT_IP, ip)

    def _get_for_user(self, action_name, user_dict):
        """Return the action matching the given user, or None."""
        self._delete_expired(action_name, time())
        for action in self._actions[action_name]:
            for action_user in action.users:
                if action_user == user_dict:
                    return action
        return None

    def get_for_user(self, user_dict):
        """Return the action matching the given user, or None."""
        return self._get_for_user(ActionName.BLOCK_USER, user_dict) \
            or self._get_for_user(ActionName.REDIRECT_USER, user_dict)

    def reload_from_dicts(self, data):
        """Reload actions from a list of dicts.

        Unsupported actions are skipped and logged but do not trigger an error.
        The list of their names is returned at the end.
        """
        self.clear()
        unsupported = []
        for action_data in data:
            try:
                action = action_from_dict(action_data)
            except UnsupportedAction:
                unsupported.append(action_data["action"])
            else:
                LOGGER.debug("Adding action %r" % action_data)
                self.add(action)
        if unsupported:
            LOGGER.error("Skipped unsupported actions: %r", unsupported)
        return unsupported


ACTION_STORE = ActionStore()



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Import hook strategy to prepare hooks for modules
"""
import logging

from ..hook_point import hook_point
from .base import BaseStrategy

LOGGER = logging.getLogger(__name__)


def parse_hook_module(hook_module):
    """ Parse a hook_module and split it into two parts

    >>> parse_hook_module('django.core.handlers.base::BaseHandler')
    ['django.core.handlers.base', 'BaseHandler']
    >>> parse_hook_module('django.core.handlers.base')
    ['django.core.handlers.base', None]
    """
    if "::" in hook_module:
        return hook_module.split("::", 1)
    else:
        return [hook_module, None]


class ImportHookStrategy(BaseStrategy):
    """ Simple strategy that calls setattr(hook_module, hook_name, callback)
    """

    def __init__(
        self,
        strategy_key,
        observation_queue,
        queue,
        import_hook,
        before_hook_point=None,
    ):
        super(ImportHookStrategy, self).__init__(
            observation_queue, queue, import_hook, before_hook_point
        )
        self.strategy_key = strategy_key

        self.hook_path = strategy_key[0]
        self.hook_module, self.hook_class = parse_hook_module(self.hook_path)
        self.hook_name = strategy_key[1]

    def hook(self):
        """ Insert the import hook
        """
        if self.hooked is False:
            self.import_hook.register_patcher(
                self.hook_module,
                self.hook_class,
                self.hook_name,
                self.import_hook_callback,
            )

            self.hooked = True
        else:
            LOGGER.warning(
                "Trying to hook several times the module %s", self.hook_module
            )

    def import_hook_callback(self, original):
        """ Monkey-patch the object located at hook_class.hook_name on an
        already loaded module. Called by ImportHook
        """
        return hook_point(self, self.hook_path, self.hook_name, original)

    @staticmethod
    def get_strategy_id(callback):
        """ Return the tuple (callback.hook_module, callback.hook_name) as
        identifier for this strategy
        """
        return (callback.hook_module, callback.hook_name)

    def _restore(self):
        """ The hooked module will always stay hooked
        """
        pass



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Look for badly behaved clients
"""
from logging import getLogger

from .regexp_rule import RegexpRule

LOGGER = getLogger(__name__)


class UserAgentMatchesCBFramework(RegexpRule):

    def pre(self, instance, args, kwargs, **options):

        request = self.storage.get_current_request()

        if not request:
            LOGGER.warning("No request was recorded abort")
            return

        user_agent = request.client_user_agent
        if not user_agent:
            return

        match = self.match_regexp(user_agent)
        if not match:
            return

        infos = {"found": match, "in": user_agent}
        self.record_attack(infos)

        return {"status": "raise", "data": match, "rule_name": self.rule_name}


UserAgentMatchesCBDjango = UserAgentMatchesCBFramework



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#

import logging

from ..ip_radix import Radix
from ..rules import RuleCallback

LOGGER = logging.getLogger(__name__)


class IPBlacklistCB(RuleCallback):

    INTERRUPTIBLE = False

    def __init__(self, *args, **kwargs):
        super(IPBlacklistCB, self).__init__(*args, **kwargs)
        self.networks = Radix(None)
        for blacklist in self.data.get("values", []):
            self.networks.insert(blacklist, '/' in blacklist)
        LOGGER.debug("Blacklisted IP networks: %s", self.networks)

    def pre(self, instance, args, kwargs, **options):
        request = self.storage.get_current_request()
        if request is None:
            return
        client_ip = request.raw_client_ip
        if client_ip is None:
            return
        network = self.networks.match(client_ip)
        if network is not None:
            LOGGER.debug(
                "IP %s belongs to blacklisted network %s",
                client_ip,
                network,
            )
            self.record_observation("blacklisted", network, 1)
            return {
                "status": "raise",
                "data": network,
                "rule_name": self.rule_name,
                "immediate": True,  # Stop execution of other callbacks on this hook point
            }



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Aggregate authentication tentatives
"""
import json
import logging

from ..utils import CustomJSONEncoder
from .matcher_callback import MatcherRule

LOGGER = logging.getLogger(__name__)


class AuthMetricsCB(MatcherRule):

    INTERRUPTIBLE = False

    def post(self, instance, args, kwargs, result=None, **options):
        # Django authentication model return either an User or None
        if result is None:
            auth_status = "auto-login-fail"
        else:
            auth_status = "auto-login-success"

        keys = []

        # Search for credentials identifier that match the whitelist
        for identifier in kwargs.keys():
            if self.match(identifier):
                value = kwargs[identifier]
                if value is not None:
                    keys.append([identifier.lower(), value])

        if not keys:
            # If we couldn't identify an user, don't record a metric
            return

        request = self.storage.get_current_request()
        if not request:
            LOGGER.warning("No request was recorded, abort")
            return
        key = {"keys": keys, "ip": request.client_ip}

        observation_key = json.dumps(
            key, separators=(",", ":"), sort_keys=True, cls=CustomJSONEncoder
        )
        self.record_observation(auth_status, observation_key, 1)



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Pyramid specific WSGI HTTP Request / Response stuff
"""
from logging import getLogger

from ..utils import is_json_serializable
from .base import BaseResponse
from .wsgi import BaseWSGIRequest

LOGGER = getLogger(__name__)


class PyramidRequest(BaseWSGIRequest):

    def __init__(self, request, storage=None):
        super(PyramidRequest, self).__init__(storage=storage)
        self.request = request

        # Cache for params
        self._query_params = None
        self._form_params = None
        self._cookies_params = None
        self._json_params = None

    @property
    def query_params(self):
        if self._query_params is None:
            try:
                # Convert pyramid MultiDict to a normal dict with values as list
                self._query_params = dict(self.request.GET.dict_of_lists())
            except Exception:
                LOGGER.debug("couldn't get request.GET from the framework",
                             exc_info=True)
                self._query_params = super(PyramidRequest, self).query_params
        return self._query_params

    @property
    def body(self):
        try:
            return self.request.body
        except Exception:
            LOGGER.debug("couldn't get request.body from the framework",
                         exc_info=True)
            return super(PyramidRequest, self).body

    @property
    def form_params(self):
        if self._form_params is None:
            try:
                # Convert pyramid MultiDict to a normal dict with values as list
                form_params = {}
                post_params = self.request.POST
                for param_name in post_params:
                    values = post_params.getall(param_name)
                    # Ignore any non json serializable value as we don't know
                    # how to process them (like cgi.FieldStorage)
                    form_params[param_name] = list(
                        filter(is_json_serializable, values)
                    )
                self._form_params = form_params
            except Exception:
                LOGGER.debug("couldn't get request.POST from framework",
                             exc_info=True)
                self._form_params = super(PyramidRequest, self).form_params
        return self._form_params

    @property
    def cookies_params(self):
        if self._cookies_params is None:
            try:
                self._cookies_params = dict(self.request.cookies)
            except Exception:
                LOGGER.debug("couldn't get request.cookies from framework",
                             exc_info=True)
                self._cookies_params = super(PyramidRequest, self).cookies_params
        return self._cookies_params

    @property
    def remote_addr(self):
        """Remote IP address."""
        return self.request.remote_addr

    @property
    def hostname(self):
        return self.request.host

    @property
    def method(self):
        return self.request.method

    @property
    def referer(self):
        return self.get_raw_header("HTTP_REFERER")

    @property
    def client_user_agent(self):
        return self.request.user_agent

    @property
    def route(self):
        """Request route."""
        route = getattr(self.request, "matched_route", None)
        pattern = getattr(route, "pattern", None)
        return pattern

    @property
    def path(self):
        return self.request.path

    @property
    def scheme(self):
        return self.request.scheme

    @property
    def server_port(self):
        return self.get_raw_header("SERVER_PORT")

    @property
    def remote_port(self):
        return self.get_raw_header("REMOTE_PORT")

    @property
    def view_params(self):
        return self.request.matchdict

    @property
    def json_params(self):
        if self._json_params is None:
            try:
                self._json_params = self.request.json_body
            except Exception:
                LOGGER.debug("couldn't get request.json_body from the framework",
                             exc_info=True)
                self._json_params = super(PyramidRequest, self).json_params
        return self._json_params

    @property
    def raw_headers(self):
        return self.request.environ


class PyramidResponse(BaseResponse):

    def __init__(self, response):
        self.response = response

    @property
    def status_code(self):
        return self.response.status_code

    @property
    def content_type(self):
        return self.response.headers.get("Content-Type")

    @property
    def content_length(self):
        try:
            return int(self.response.content_length)
        except (ValueError, TypeError):
            return None



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Look for badly behaved clients
"""
from logging import getLogger

from ..rules import RuleCallback

LOGGER = getLogger(__name__)


class NotFoundCBFlask(RuleCallback):
    def post(self, instance, args, kwargs, result=None, **options):
        response = result
        if response.status_code == 404:
            current_request = self.storage.get_current_request()

            # A 404 may prevent record request context to store the request
            # store it without arguments
            if current_request is None:
                LOGGER.warning("No request was recorded abort")
                return

            infos = {
                "path": current_request.path,
                "host": current_request.hostname,
                "verb": current_request.method,
                "ua": current_request.client_user_agent,
            }
            self.record_attack(infos)



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Module called when sitecustomize is automatically imported by python
"""
import imp
import os.path
import sys

from sqreen import start

LOADING = True

# Try to import other sitecustomize files in the path

current_directory = os.path.dirname(__file__)

# Remove ourselves from sys.path avoiding an import loop
if current_directory in sys.path:
    sys.path.remove(current_directory)

try:
    (file, pathname, description) = imp.find_module("sitecustomize", sys.path)
except ImportError:
    # There can no other sitecustomize
    pass
else:
    imp.load_module("sitecustomize", file, pathname, description)


start()



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Hijacking strategies
"""

from ...utils import HAS_ASYNCIO
from .base import BaseStrategy
from .dbapi2 import DBApi2Strategy
from .django_strategy import DjangoStrategy
from .flask_strategy import FlaskStrategy
from .import_hook import ImportHookStrategy
from .lxml_strategy import LXMLResolverStrategy
from .psycopg2_strategy import Psycopg2Strategy
from .pyramid_strategy import PyramidStrategy
from .wsgi_strategy import (
    WSGIFactoryStrategy,
    WSGIReceiverStrategy,
    WSGIStrategy,
)

if HAS_ASYNCIO:
    from .aiohttp_strategy import AioHTTPHookStrategy, AioHTTPInstallStrategy
    from .async_event_loop import AsyncEventLoopStrategy
    from .async_import_hook import AsyncImportHookStrategy

__all__ = [
    "AioHTTPHookStrategy",
    "AioHTTPInstallStrategy",
    "AsyncEventLoopStrategy",
    "AsyncImportHookStrategy",
    "BaseStrategy",
    "DBApi2Strategy",
    "DjangoStrategy",
    "FlaskStrategy",
    "LXMLResolverStrategy",
    "ImportHookStrategy",
    "Psycopg2Strategy",
    "PyramidStrategy",
    "WSGIFactoryStrategy",
    "WSGIReceiverStrategy",
    "WSGIStrategy"
]



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Base strategy
"""
import logging
from bisect import bisect
from collections import defaultdict

from ..hook_point import LIFECYCLE_METHODS

LOGGER = logging.getLogger(__name__)


class BaseStrategy(object):
    """ The base strategy holds callbacks for a hook point

    Subclasses need to define way to hook and deinstrument.
    BaseStrategy accepts a channel, usually passed from Instrumentation
    directly.
    """

    def __init__(
        self, observation_queue, queue, import_hook, before_hook_point=None
    ):
        self.observation_queue = observation_queue
        self.queue = queue
        self.unique_callbacks = set()
        # Data structure for {[hook_module, hook_name]: {pre: [], post: [], fail: []}}
        self.callbacks = defaultdict(lambda: defaultdict(list))
        self.import_hook = import_hook

        self.hooked = False

        self._before_hook_point = before_hook_point

    def add_callback(self, callback):
        """ Add a callback on a specific endpoint defined by the couple
        (hook_module, hook_name) embedded in callback
        """
        key = (callback.hook_module, callback.hook_name)

        # Normal order insertion
        if hasattr(callback, LIFECYCLE_METHODS["PRE"]):
            callbacks = self.callbacks[key][LIFECYCLE_METHODS["PRE"]]
            idx = bisect([c.priority for c in callbacks], callback.priority)
            callbacks.insert(idx, callback)

        # Reverse order insertion
        for method in (LIFECYCLE_METHODS["POST"], LIFECYCLE_METHODS["FAILING"]):
            if hasattr(callback, method):
                rev_callbacks = list(reversed(self.callbacks[key][method]))
                idx = bisect([c.priority for c in rev_callbacks], callback.priority)
                rev_callbacks.insert(idx, callback)
                rev_callbacks.reverse()
                self.callbacks[key][method] = rev_callbacks

        self.unique_callbacks.add(callback)

    def hook(self):
        pass

    def get_pre_callbacks(self, key):
        """ Returns callbacks with a pre lifecycle method for the key
        (hook_module, hook_name)
        """
        return self.callbacks[key][LIFECYCLE_METHODS["PRE"]]

    def get_post_callbacks(self, key):
        """ Returns callbacks with a post lifecycle method for the key
        (hook_module, hook_name)
        """
        return self.callbacks[key][LIFECYCLE_METHODS["POST"]]

    def get_failing_callbacks(self, key):
        """ Returns callbacks with a failing lifecycle method for the key
        (hook_module, hook_name)
        """
        return self.callbacks[key][LIFECYCLE_METHODS["FAILING"]]

    def _total_callbacks(self):
        """ Count the total number of callbacks
        """
        return len(self.unique_callbacks)

    def deinstrument(self, callback):
        """ Deactivate the callback for this endpoint.
        If it was the latests callback for this endpoint, restore the original
        hooked function.
        """
        key = (callback.hook_module, callback.hook_name)

        callbacks = self.callbacks[key]

        LOGGER.debug("Deinstrumenting %s", callback)

        for method in LIFECYCLE_METHODS.values():
            if hasattr(callback, method):
                try:
                    callbacks[method].remove(callback)
                except ValueError:
                    # Callback has not been added in this strategy
                    msg = "Callback %s %s trying to be deinstrumented but not instrumented"
                    LOGGER.warning(msg, method, callback)

        try:
            self.unique_callbacks.remove(callback)
        except KeyError:
            pass

        LOGGER.debug(
            "Number of remaining callback for %s: %s",
            self,
            self._total_callbacks(),
        )

        # If no more callbacks are set deinstrument
        if self._total_callbacks() == 0:
            self._restore()

    def deinstrument_all(self):
        """ Deinstrument all callbacks
        """
        for callback in list(self.unique_callbacks):
            self.deinstrument(callback)

        self.unique_callbacks = set()

    def _restore(self):
        """ Restore the original method, subclasses need to overload this method
        """
        self.hooked = False

    def before_hook_point(self):
        """ Run code just before running a hook_point
        """
        if self._before_hook_point is not None:
            self._before_hook_point()



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
from ...frameworks.django_framework import DjangoRequest
from ...runtime_storage import runtime
from .base import BaseMiddleware


class DjangoMiddleware(BaseMiddleware):
    """ Wrap a RuleCallback and alias its methods to django middleware methods.
    Pre is mapped to process_view, post is mapped to process_reponse and
    failig is mapped to process_exception
    """

    def process_view(self, request, view_func, view_args, view_kwargs):
        """ Call wrapped_callback.pre, raise AttackBlock if needed.
        Django pre callbacks will receive these arguments:
        (None, request, view_func, view_args, view_kwargs)
        """
        self.strategy.before_hook_point()
        args = (request, view_func, view_args, view_kwargs)

        self.execute_pre_callbacks(args)

    def process_response(self, request, response):
        """ Call wrapped_callback.post, raise AttackBlock if needed or returns
        the response passed as input.
        Django post callbacks will receive these arguments:
        (None, response, request)
        """
        # Record request if we don't have one yet
        runtime.store_request_default(DjangoRequest(request))

        # Execute pre_callbacks if the response is a 404 because the callback
        # process_view was not called by Django
        if response.status_code == 404:
            args = (request, None, None, None)
            self.execute_pre_callbacks(args)
        args = (request,)
        return self.execute_post_callbacks(response, args)

    def process_exception(self, request, exception):
        """ Call wrapped_callback.failing, always return None.
        Django failing callbacks will receive these arguments:
        (None, exception, request)
        """
        args = (request,)

        # Record request if we don't have one yet
        runtime.store_request_default(DjangoRequest(request))

        self.execute_failing_callbacks(exception, args)



# -*- coding: utf-8 -*-
#
#  Copyright 2011 Sybren A. Stüvel <sybren@stuvel.eu>
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Deprecated version of the RSA module

.. deprecated:: 2.0

    This submodule is deprecated and will be completely removed as of version 4.0.

Module for calculating large primes, and RSA encryption, decryption,
signing and verification. Includes generating public and private keys.

WARNING: this code implements the mathematics of RSA. It is not suitable for
real-world secure cryptography purposes. It has not been reviewed by a security
expert. It does not include padding of data. There are many ways in which the
output of this module, when used without any modification, can be sucessfully
attacked.
"""

__author__ = "Sybren Stuvel, Marloes de Boer and Ivo Tamboer"
__date__ = "2010-02-05"
__version__ = '1.3.3'

# NOTE: Python's modulo can return negative numbers. We compensate for
# this behaviour using the abs() function

try:
    import cPickle as pickle
except ImportError:
    import pickle
from pickle import dumps, loads
import base64
import math
import os
import random
import sys
import types
import zlib

from ._compat import byte

# Display a warning that this insecure version is imported.
import warnings
warnings.warn('Insecure version of the RSA module is imported as %s, be careful'
        % __name__)
warnings.warn('This submodule is deprecated and will be completely removed as of version 4.0.',
              DeprecationWarning)


def gcd(p, q):
    """Returns the greatest common divisor of p and q


    >>> gcd(42, 6)
    6
    """
    if p<q: return gcd(q, p)
    if q == 0: return p
    return gcd(q, abs(p%q))

def bytes2int(bytes):
    """Converts a list of bytes or a string to an integer
    """

    if not (type(bytes) is types.ListType or type(bytes) is types.StringType):
        raise TypeError("You must pass a string or a list")

    # Convert byte stream to integer
    integer = 0
    for byte in bytes:
        integer *= 256
        if type(byte) is types.StringType: byte = ord(byte)
        integer += byte

    return integer

def int2bytes(number):
    """Converts a number to a string of bytes
    """

    if not (type(number) is types.LongType or type(number) is types.IntType):
        raise TypeError("You must pass a long or an int")

    string = ""

    while number > 0:
        string = "%s%s" % (byte(number & 0xFF), string)
        number /= 256
    
    return string

def fast_exponentiation(a, p, n):
    """Calculates r = a^p mod n
    """
    result = a % n
    remainders = []
    while p != 1:
        remainders.append(p & 1)
        p = p >> 1
    while remainders:
        rem = remainders.pop()
        result = ((a ** rem) * result ** 2) % n
    return result

def read_random_int(nbits):
    """Reads a random integer of approximately nbits bits rounded up
    to whole bytes"""

    nbytes = ceil(nbits/8.)
    randomdata = os.urandom(nbytes)
    return bytes2int(randomdata)

def ceil(x):
    """ceil(x) -> int(math.ceil(x))"""

    return int(math.ceil(x))
    
def randint(minvalue, maxvalue):
    """Returns a random integer x with minvalue <= x <= maxvalue"""

    # Safety - get a lot of random data even if the range is fairly
    # small
    min_nbits = 32

    # The range of the random numbers we need to generate
    range = maxvalue - minvalue

    # Which is this number of bytes
    rangebytes = ceil(math.log(range, 2) / 8.)

    # Convert to bits, but make sure it's always at least min_nbits*2
    rangebits = max(rangebytes * 8, min_nbits * 2)
    
    # Take a random number of bits between min_nbits and rangebits
    nbits = random.randint(min_nbits, rangebits)
    
    return (read_random_int(nbits) % range) + minvalue

def fermat_little_theorem(p):
    """Returns 1 if p may be prime, and something else if p definitely
    is not prime"""

    a = randint(1, p-1)
    return fast_exponentiation(a, p-1, p)

def jacobi(a, b):
    """Calculates the value of the Jacobi symbol (a/b)
    """

    if a % b == 0:
        return 0
    result = 1
    while a > 1:
        if a & 1:
            if ((a-1)*(b-1) >> 2) & 1:
                result = -result
            b, a = a, b % a
        else:
            if ((b ** 2 - 1) >> 3) & 1:
                result = -result
            a = a >> 1
    return result

def jacobi_witness(x, n):
    """Returns False if n is an Euler pseudo-prime with base x, and
    True otherwise.
    """

    j = jacobi(x, n) % n
    f = fast_exponentiation(x, (n-1)/2, n)

    if j == f: return False
    return True

def randomized_primality_testing(n, k):
    """Calculates whether n is composite (which is always correct) or
    prime (which is incorrect with error probability 2**-k)

    Returns False if the number if composite, and True if it's
    probably prime.
    """

    q = 0.5     # Property of the jacobi_witness function

    # t = int(math.ceil(k / math.log(1/q, 2)))
    t = ceil(k / math.log(1/q, 2))
    for i in range(t+1):
        x = randint(1, n-1)
        if jacobi_witness(x, n): return False
    
    return True

def is_prime(number):
    """Returns True if the number is prime, and False otherwise.
    """

    """
    if not fermat_little_theorem(number) == 1:
        # Not prime, according to Fermat's little theorem
        return False
    """

    if randomized_primality_testing(number, 5):
        # Prime, according to Jacobi
        return True
    
    # Not prime
    return False

    
def getprime(nbits):
    """Returns a prime number of max. 'math.ceil(nbits/8)*8' bits. In
    other words: nbits is rounded up to whole bytes.
    """

    nbytes = int(math.ceil(nbits/8.))

    while True:
        integer = read_random_int(nbits)

        # Make sure it's odd
        integer |= 1

        # Test for primeness
        if is_prime(integer): break

        # Retry if not prime

    return integer

def are_relatively_prime(a, b):
    """Returns True if a and b are relatively prime, and False if they
    are not.
    """

    d = gcd(a, b)
    return (d == 1)

def find_p_q(nbits):
    """Returns a tuple of two different primes of nbits bits"""

    p = getprime(nbits)
    while True:
        q = getprime(nbits)
        if not q == p: break
    
    return (p, q)

def extended_euclid_gcd(a, b):
    """Returns a tuple (d, i, j) such that d = gcd(a, b) = ia + jb
    """

    if b == 0:
        return (a, 1, 0)

    q = abs(a % b)
    r = long(a / b)
    (d, k, l) = extended_euclid_gcd(b, q)

    return (d, l, k - l*r)

# Main function: calculate encryption and decryption keys
def calculate_keys(p, q, nbits):
    """Calculates an encryption and a decryption key for p and q, and
    returns them as a tuple (e, d)"""

    n = p * q
    phi_n = (p-1) * (q-1)

    while True:
        # Make sure e has enough bits so we ensure "wrapping" through
        # modulo n
        e = getprime(max(8, nbits/2))
        if are_relatively_prime(e, n) and are_relatively_prime(e, phi_n): break

    (d, i, j) = extended_euclid_gcd(e, phi_n)

    if not d == 1:
        raise Exception("e (%d) and phi_n (%d) are not relatively prime" % (e, phi_n))

    if not (e * i) % phi_n == 1:
        raise Exception("e (%d) and i (%d) are not mult. inv. modulo phi_n (%d)" % (e, i, phi_n))

    return (e, i)


def gen_keys(nbits):
    """Generate RSA keys of nbits bits. Returns (p, q, e, d).

    Note: this can take a long time, depending on the key size.
    """

    while True:
        (p, q) = find_p_q(nbits)
        (e, d) = calculate_keys(p, q, nbits)

        # For some reason, d is sometimes negative. We don't know how
        # to fix it (yet), so we keep trying until everything is shiny
        if d > 0: break

    return (p, q, e, d)

def gen_pubpriv_keys(nbits):
    """Generates public and private keys, and returns them as (pub,
    priv).

    The public key consists of a dict {e: ..., , n: ....). The private
    key consists of a dict {d: ...., p: ...., q: ....).
    """
    
    (p, q, e, d) = gen_keys(nbits)

    return ( {'e': e, 'n': p*q}, {'d': d, 'p': p, 'q': q} )

def encrypt_int(message, ekey, n):
    """Encrypts a message using encryption key 'ekey', working modulo
    n"""

    if type(message) is types.IntType:
        return encrypt_int(long(message), ekey, n)

    if not type(message) is types.LongType:
        raise TypeError("You must pass a long or an int")

    if message > 0 and \
            math.floor(math.log(message, 2)) > math.floor(math.log(n, 2)):
        raise OverflowError("The message is too long")

    return fast_exponentiation(message, ekey, n)

def decrypt_int(cyphertext, dkey, n):
    """Decrypts a cypher text using the decryption key 'dkey', working
    modulo n"""

    return encrypt_int(cyphertext, dkey, n)

def sign_int(message, dkey, n):
    """Signs 'message' using key 'dkey', working modulo n"""

    return decrypt_int(message, dkey, n)

def verify_int(signed, ekey, n):
    """verifies 'signed' using key 'ekey', working modulo n"""

    return encrypt_int(signed, ekey, n)

def picklechops(chops):
    """Pickles and base64encodes it's argument chops"""

    value = zlib.compress(dumps(chops))
    encoded = base64.encodestring(value)
    return encoded.strip()

def unpicklechops(string):
    """base64decodes and unpickes it's argument string into chops"""

    return loads(zlib.decompress(base64.decodestring(string)))

def chopstring(message, key, n, funcref):
    """Splits 'message' into chops that are at most as long as n,
    converts these into integers, and calls funcref(integer, key, n)
    for each chop.

    Used by 'encrypt' and 'sign'.
    """

    msglen = len(message)
    mbits = msglen * 8
    nbits = int(math.floor(math.log(n, 2)))
    nbytes = nbits / 8
    blocks = msglen / nbytes

    if msglen % nbytes > 0:
        blocks += 1

    cypher = []
    
    for bindex in range(blocks):
        offset = bindex * nbytes
        block = message[offset:offset+nbytes]
        value = bytes2int(block)
        cypher.append(funcref(value, key, n))

    return picklechops(cypher)

def gluechops(chops, key, n, funcref):
    """Glues chops back together into a string.  calls
    funcref(integer, key, n) for each chop.

    Used by 'decrypt' and 'verify'.
    """
    message = ""

    chops = unpicklechops(chops)
    
    for cpart in chops:
        mpart = funcref(cpart, key, n)
        message += int2bytes(mpart)
    
    return message

def encrypt(message, key):
    """Encrypts a string 'message' with the public key 'key'"""
    
    return chopstring(message, key['e'], key['n'], encrypt_int)

def sign(message, key):
    """Signs a string 'message' with the private key 'key'"""
    
    return chopstring(message, key['d'], key['p']*key['q'], decrypt_int)

def decrypt(cypher, key):
    """Decrypts a cypher with the private key 'key'"""

    return gluechops(cypher, key['d'], key['p']*key['q'], decrypt_int)

def verify(cypher, key):
    """Verifies a cypher with the public key 'key'"""

    return gluechops(cypher, key['e'], key['n'], encrypt_int)

# Do doctest if we're not imported
if __name__ == "__main__":
    import doctest
    doctest.testmod()

__all__ = ["gen_pubpriv_keys", "encrypt", "decrypt", "sign", "verify"]




# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Count things according to a list of binding accessors expression
"""
import json
import logging

from ..binding_accessor import BindingAccessor
from ..exceptions import InvalidArgument
from ..frameworks.wsgi import WSGIResponse
from ..rules import RuleCallback
from ..utils import Iterable

LOGGER = logging.getLogger(__name__)


class BindingAccessorCounter(RuleCallback):
    def __init__(self, *args, **kwargs):
        super(BindingAccessorCounter, self).__init__(*args, **kwargs)

        values = self.data.get("values")
        if values is None or not isinstance(values, Iterable):
            msg = "Invalid data type received: {}"
            raise InvalidArgument(msg.format(type(self.data)))

        self.patterns = [BindingAccessor(exp) for exp in values]
        self.metric_name = self.metrics[0]["name"]

    def post(self, instance, args, kwargs, result=None, **options):
        """ Resolve binding expressions with the HTTP Response and send
        an observation for the rule metric.
        """
        request = self.storage.get_current_request()
        response = self.storage.get_current_response()

        binding_eval_args = {
            "request": request,
            "response": response,
            "inst": instance,
            "args": self.storage.get_current_args(args),
            "kwargs": kwargs,
            "data": self.data,
            "rv": result,
        }

        key = [
            binding.resolve(**binding_eval_args) for binding in self.patterns
        ]
        formatted_key = json.dumps(key, separators=(",", ":"))
        self.record_observation(self.metric_name, formatted_key, 1)

    def failing(self, instance, args, kwargs, result_action=None, **options):
        """ Resolve binding expressions with the exception, get the status code
        from the result action and send an observation for the rule metric.
        """
        request = self.storage.get_current_request()
        response = self.storage.get_current_response()
        result = response or WSGIResponse("500 Internal Server Error")
        if result_action is not None \
                and result_action.get("status") == "override":
            result = result_action.get("new_return_value")

        binding_eval_args = {
            "request": request,
            "response": response,
            "inst": instance,
            "args": self.storage.get_current_args(args),
            "kwargs": kwargs,
            "data": self.data,
            "rv": result,
        }

        key = [
            binding.resolve(**binding_eval_args) for binding in self.patterns
        ]
        formatted_key = json.dumps(key, separators=(",", ":"))
        self.record_observation(self.metric_name, formatted_key, 1)
